
from importlib import resources as impresources
import networkx as nx
import numpy as np
from pydrake.all import (
    AbstractValue,
    CollisionFilterDeclaration,
    ConstantValueSource,
    ConstantVectorSource,
    Context,
    Diagram,
    GeometrySet,
    InverseKinematics,
    MultibodyPlant,
    Parser,
    Quaternion,
    RigidTransform,
    RotationMatrix,
)
from pydrake.math import RollPitchYaw, RigidTransform
from typing import Callable, List, Tuple

# Internal Imports
from brom_drake.directories import DEFAULT_BROM_MODELS_DIR
from brom_drake.control import (
    GripperTarget,
)
from brom_drake.file_manipulation.urdf import drakeify_my_urdf
from brom_drake.file_manipulation.urdf.shapes.box import BoxDefinition
from brom_drake.file_manipulation.urdf.simple_writer.urdf_definition import SimpleShapeURDFDefinition, \
    InertiaDefinition
from brom_drake.motion_planning.systems.open_loop_dispensers.open_loop_plan_dispenser import OpenLoopPlanDispenser
import brom_drake.robots as robots
from brom_drake.robots.gripper_type import GripperType
from brom_drake.stations.classical import UR10eStation
from brom_drake.productions import ProductionID
from brom_drake.productions.roles import Role
from brom_drake.productions.types import OfflineDynamicMotionPlanningProduction
from brom_drake.utils import Performer, AddGround, MotionPlan


class ChemLab2(OfflineDynamicMotionPlanningProduction):
    """
    *Description*

    This production is the second in the chemistry lab series.
    It is used to test the motion planning capabilities of the robot
    in a chemistry lab setting with minimal constraints, but with full
    dynamics of the robot included in the simulation.

    The earlier chemistry lab production (ChemLab1) ignored the physics of the
    arm and objects in order to simplify some calculations.

    *Parameters*

    time_step: float, optional
        The time step for the production, by default 1e-3.

    meshcat_port_number: int, optional
        The port number for meshcat visualization, by default 7001.

    plan_execution_speed: float, optional
        The speed at which to execute the motion plan, by default 0.2.

    pose_WorldBeaker: RigidTransform, optional
        The pose of the beaker in the world frame, by default None.

    table_length: float, optional
        The length of the table, by default 0.6.

    table_width: float, optional
        The width of the table, by default 2.0.

    table_height: float, optional
        The height of the table, by default 0.1.

    shelf_pose: RigidTransform, optional
        The pose of the shelf in the world frame, by default None.

    gripper_type: GripperType, optional
        The type of gripper to use, by default GripperType.Robotiq_2f_85.

    *Usage*

    To use this production, create an instance of the ChemLab2 class,
    provide it with a motion planning function, and then build it. ::

        from brom_drake.motion_planning.algorithms.rrt.bidirectional_connect import (
            BidirectionalRRTConnectPlanner,
            BidirectionalRRTConnectPlannerConfig,
        )
        from brom_drake.productions import ChemLab2
    
        # Create the production
        production = ChemLab2()

        # Create a planner object which will be used to plan the motion
        config = BidirectionalRRTConnectPlannerConfig(
            steering_step_size=0.2,
        )
        planner2 = BidirectionalRRTConnectPlanner(
            production.arm,
            production.plant,
            production.scene_graph,
            config=config,
        )

        # To build the production, we only need to provide a planning function
        # (can come from anywhere, not just a BaseRRTPlanner object)
        diagram, diagram_context = production.easy_cast_and_build(
            planner2.plan,
            with_watcher=True,
        )

        print("Simulating...")

        # Simulate the diagram
        simulator = Simulator(diagram, diagram_context)
        simulator.set_target_realtime_rate(2.0)

        # Run simulation
        simulator.Initialize()
        simulator.AdvanceTo(0.05)
        planned_trajectory = production.plan_dispenser.planned_trajectory

        print(f"Expected end time of trajectory: {planned_trajectory.end_time()}")
        simulator.AdvanceTo(planned_trajectory.end_time()+2.0)

    """
    def __init__(
        self,
        time_step: float = 1e-3,
        meshcat_port_number: int = 7001,
        plan_execution_speed: float = 0.2,
        pose_WorldBeaker: RigidTransform = None,
        table_length: float = 0.6,
        table_width: float = 2.0,
        table_height: float = 0.1,
        shelf_pose: RigidTransform = None,
        gripper_type: GripperType = GripperType.Robotiq_2f_85,
        **kwargs,
    ):
        """
        *Description*
            
        Constructor for the ChemLab2 Production.

        *Parameters*

        time_step: float, optional
            The time step for the production, by default 1e-3.

        meshcat_port_number: int, optional
            The port number for meshcat visualization, by default 7001.

        plan_execution_speed: float, optional
            The speed at which to execute the motion plan, by default 0.2.

        pose_WorldBeaker: RigidTransform, optional
            The pose of the beaker in the world frame, by default None.

        table_length: float, optional
            The length of the table, by default 0.6.

        table_width: float, optional
            The width of the table, by default 2.0.

        table_height: float, optional
            The height of the table, by default 0.1.

        shelf_pose: RigidTransform, optional
            The pose of the shelf in the world frame, by default None.

        gripper_type: GripperType, optional
            The type of gripper to use, by default GripperType.Robotiq_2f_85.
        
        """
        # Superclass constructor
        super().__init__(**kwargs)

        # Input Processing
        self.time_step = time_step
        self.meshcat_port_number = meshcat_port_number
        self.plan_execution_speed = plan_execution_speed
        self.table_height, self.table_width = table_height, table_width
        self.table_length = table_length
        self.gripper_type = gripper_type

        # Initialize pose data
        self.shelf_pose = shelf_pose
        self.pose_WorldBeaker = pose_WorldBeaker
        self.pose_BeakerGoal = None
        self.initialize_pose_data()

        # Set station
        self.station = UR10eStation(
            time_step=self.time_step,
            meshcat_port_number=self.meshcat_port_number,
            gripper_type=self.gripper_type,
        )
        self.arm = self.station.arm
        self.robot_model_idx_ = self.arm

        # self.gripper = self.station.gripper

        # Set Names of Plant and scene graph
        self.plant = self.station.plant
        self.scene_graph = self.station.scene_graph

        self.plant.set_name(f"ChemLab2_Production_Plant")
        self.scene_graph.set_name(f"ChemLab2_Production_SceneGraph")

        # Define placeholder variables for models
        self.test_tube_holder1 = None

    def add_supporting_cast(self):
        """
        *Description*
        
        This method adds all secondary cast members to the builder.
        The secondary cast members in the production are the:

        - Table, where the robot exists
        - Test Tube Holders
        - Component which share's the robot model reference
        - Motion Planning components (e.g., dispensers, etc.)
        - Start and Goal sources
        """
        # Call the superclass method
        super().add_supporting_cast()

        # Setup

        # Create the table and the items on it
        self.add_table()
        self.add_test_tube_holders()
        self.add_beaker()
        self.add_shelf()
        AddGround(self.plant)

        # Add the station
        self.builder.AddSystem(
            self.station,
        )

        self.station.Finalize()

        # Add the motion planning components
        self.add_robot_source_system()
        self.add_motion_planning_components()
        self.add_start_and_goal_sources_to_builder()

        # Add initial conditions for the arm
        self.initial_condition_manager.add_initial_configuration(
            model_instance_index=self.arm,
            configuration=self.start_configuration,
        )
        # The initial condition will be set for the plant when we build the production.

        print("Completed adding supporting cast members.")

    def add_beaker(self):
        """
        *Description*
        
        This method adds the beaker URDF to the production's plant.
        The beaker's URDF comes from the brom_drake `robots` directories.
        """
        # Setup
        plant = self.plant
        urdf_file_path = str(
            impresources.files(robots) / "models/beaker/beaker.urdf"
        )

        # Use Drakeify my urdf to create the beaker
        new_beaker_urdf = drakeify_my_urdf(urdf_file_path)

        # Add the beaker to the plant
        model_idcs = Parser(plant=plant).AddModels(str(new_beaker_urdf))
        self.beaker_model_index = model_idcs[0]

        # Add object to the list we use for initialization
        self.models_in_supporting_cast.append(
            (self.beaker_model_index, self.pose_WorldBeaker),
        )

        # # Weld the beaker to the world frame
        # plant.WeldFrames(
        #     plant.world_frame(),
        #     plant.GetFrameByName("beaker_base_link", self.beaker_model_index),
        #     self.pose_WorldBeaker,
        # )

    def add_motion_planning_components(self):
        """
        *Description*

        Add the following components to the production's
        ``builder`` in order to help with motion planning:

        - `plan_dispenser`: An OpenLoopPlanDispenser which will describe the reference robot trajectory.
        - Connections between reference trajectory and UR10e station.
        - Gripper control components (type of control and an "open" command)
        """
        # Setup
        arm = self.arm
        n_actuated_dof = self.plant.num_actuated_dofs(arm) # Number of actuated DOF in arm

        # Add the Plan Dispenser and connect it to the station
        self.plan_dispenser = self.builder.AddSystem(
            OpenLoopPlanDispenser(n_actuated_dof, self.plan_execution_speed)
        )

        self.builder.Connect(
            self.plan_dispenser.GetOutputPort("point_in_plan"),
            self.station.GetInputPort("desired_joint_positions"),
        )

        if self.gripper_type != GripperType.NoGripper:
            # Add components for the gripper's control, including:
            # - Gripper Target Type
            # - Gripper Target
            # Both should be static.
            
            # Add the gripper target to the builder
            gripper_target_type_source = self.builder.AddSystem(
                ConstantValueSource(
                    AbstractValue.Make(GripperTarget.kPosition),
                ),
            )

            # Connect the gripper target type to the station
            self.builder.Connect(
                gripper_target_type_source.get_output_port(),
                self.station.GetInputPort("gripper_target_type"),
            )

            # Add the gripper target to the builder
            gripper_target_source = self.builder.AddSystem(
                ConstantVectorSource(np.array([0.0])),
            )

            # Connect the gripper target to the station
            self.builder.Connect(
                gripper_target_source.get_output_port(),
                self.station.GetInputPort("gripper_target"),
            )


    def add_shelf(self):
        """
        *Description*

        Add the shelf to the production's plant.
        Welds the shelf to the world frame.
        """
        # Setup
        plant = self.plant
        urdf_file_path = str(
            impresources.files(robots) / "models/cupboard/cupboard.sdf"
        )

        # Add the shelf to the plant
        model_idcs = Parser(plant=plant).AddModels(urdf_file_path)
        self.shelf_model_index = model_idcs[0]

        # Weld the bookshelf to the world frame
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("cupboard_body", self.shelf_model_index),
            self.pose_WorldShelf,
        )

    def add_table(self):
        """
        *Description*
        
        This method adds the table to the production's plant.
        Welds the table to the world frame.
        """
        # Setup

        # Create a new shape urdf
        table_defn = SimpleShapeURDFDefinition(
            name="table",
            shape=BoxDefinition(
                size=(self.table_width, self.table_length, self.table_height),
            ),
            mass=100.0, # kg
            inertia=InertiaDefinition(
                ixx=10.0,
                iyy=10.0,
                izz=10.0,
            ),
            color=np.array([0.1, 0.1, 0.1, 1.0]),
        )
        table_urdf_path = DEFAULT_BROM_MODELS_DIR + "/table/table.urdf"
        table_defn.write_to_file(table_urdf_path)

        # Add the table to the production
        table_model_index = Parser(self.plant).AddModels(table_urdf_path)[0]

        # Weld the table to the world frame
        table_pose = RigidTransform(
            RollPitchYaw(0.0, 0.0, 0.0).ToQuaternion(),
            np.array([0.0, 0.6, 0.0]),
        )
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName("table_base_link", table_model_index),
            table_pose,
        )

    def add_test_tube_holders(self):
        """
        *Description*
        
        This method adds the test tube holders to the production's plant.
        Welds the test tube holders to the world frame with using
        predefined poses.
        """
        # Setup
        test_tube_urdf1 = str(
            impresources.files(robots) / "models/test_tube_holder/test_tube_holder.urdf"
        )

        # Use Drakeify my urdf to create the test tube holders
        new_test_tube_holder_urdf1 = drakeify_my_urdf(test_tube_urdf1)

        self.test_tube_holder1 = Parser(self.plant).AddModels(
            str(new_test_tube_holder_urdf1)
        )[0]

        # Add object to the list we use for initialization
        self.models_in_supporting_cast.append(
            (self.test_tube_holder1, self.pose_WorldHolder),
        )

    def add_cast_and_build(
        self,
        cast: List[Tuple[Role, Performer]] = [],
        with_watcher: bool = False,
    ) -> Tuple[Diagram, Context]:
        """
        *Description*
        
        Modifies the normal add_cast_and_build, so that
        we share the context of the plant with the appropriate
        parts of the system.

        TODO(kwesi): Consider making this a standard part of the
        OfflineDynamicMotionPlanningProduction class.

        *Parameters*

        cast: List[Tuple[Role, Performer]], optional
            The cast to add to the production, by default [].

        with_watcher: bool, optional
            A Boolean that determines whether to add a watcher to the diagram, by default False.

        *Returns*

        diagram: Diagram
            The built diagram.

        diagram_context: Context
            The context for the built diagram.
            This context will be used to initialize the watcher.
        """
        diagram, diagram_context = super().add_cast_and_build(
            main_cast_members=cast,
            with_watcher=with_watcher,
        )

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
            self.station.arm_controller.plant, diagram_context,
        )

        for role_ii, performer_ii in cast:
            if role_ii.name == "OfflineMotionPlanner":
                performer_ii.set_internal_root_context(
                    diagram_context
                )

        return diagram, diagram_context
    
    def configure_collision_filter(self, scene_graph_context: Context):
        """
        *Description*
        
        This method configures the collision filter in the scene graph so that we ignore:
        - Self collisions of the shelf
        - Collisions between adjacent robot links (TODO)

        *Parameters*

        scene_graph_context: Context
            The context for the scene graph.
        """
        # Setup
        scene_graph = self.scene_graph
        shelf_model_index = self.shelf_model_index

        # Ignore self collisions of the shelf, by:
        # - Getting the model instance index of the shelf
        # - Collecting the Geometry IDs of all the geometries in the shelf
        shelf_geometry_ids = []
        for body_index in self.plant.GetBodyIndices(shelf_model_index):
            # Get the geometry IDs
            shelf_geometry_ids.extend(
                self.plant.GetCollisionGeometriesForBody(
                    self.plant.get_body(body_index)
                )
            )
            # print(self.plant.get_body(body_index))

        self.shelf_geometry_ids = shelf_geometry_ids

        # Apply the collision filter
        scene_graph.collision_filter_manager(scene_graph_context).Apply(
            CollisionFilterDeclaration().ExcludeWithin(
                GeometrySet(self.shelf_geometry_ids)
            )
        )

        # # Ignore collisions between the robot links
        # # TODO: Actually implement this for adjacent links? Or is this not necessary?
        # arm_geometry_ids = []
        # for body_index in self.plant.GetBodyIndices(self.arm):
        #     arm_geometry_ids.extend(
        #         self.plant.GetCollisionGeometriesForBody(
        #             self.plant.get_body(body_index)
        #         )
        #     )

        # self.arm_geometry_ids = arm_geometry_ids

        # # Apply the collision filter
        # scene_graph.collision_filter_manager(scene_graph_context).Apply(
        #     CollisionFilterDeclaration().ExcludeWithin(
        #         GeometrySet(self.arm_geometry_ids)
        #     )
        # )
    
    def define_pose_ik_problem(
        self,
        pose_WorldTarget: RigidTransform,
        target_frame_name: str,
        eps0: float = 2.5e-2,
        orientation_cost_scaling: float = 0.25,
    ) -> InverseKinematics:
        """
        *Description*
        
        Sets up the inverse kinematics problem for the start pose
        input to theis function.
        :return:
        """
        # Setup
        scene_graph = self.scene_graph
        sg_inspector = scene_graph.model_inspector()

        # Create IK Problem
        ik_problem = InverseKinematics(self.plant)

        # Add Pose Target
        ik_problem.AddPositionConstraint(
            self.plant.world_frame(),
            pose_WorldTarget.translation(),
            self.plant.GetFrameByName(target_frame_name),
            (- np.ones((3,)) * eps0).reshape((-1, 1)),
            (+ np.ones((3,)) * eps0).reshape((-1, 1)),
        )

        # TODO(kwesi): Add OrientationCosntraint
        # ik_problem.AddOrientationConstraint(
        #     self.plant.world_frame(),
        #     RotationMatrix(Quaternion(input_pose_vec[3:]).rotation()),
        #     self.plant.GetFrameByName("ft_frame"),
        #     RotationMatrix.Identity(),
        #     0.25,
        # )

        ik_problem.AddOrientationConstraint(
            self.plant.world_frame(),
            pose_WorldTarget.rotation(),
            self.plant.GetFrameByName(target_frame_name),
            RotationMatrix.Identity(),
            np.pi/8.0
        )

        # ik_problem.AddOrientationCost(
        #     self.plant.world_frame(),
        #     pose_WorldTarget.rotation(),
        #     self.plant.GetFrameByName(target_frame_name),
        #     RotationMatrix.Identity(),
        #     orientation_cost_scaling,
        # )

        return ik_problem

    def easy_cast_and_build(
        self,
        planning_algorithm: Callable[
            [np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],
            Tuple[MotionPlan, bool],
        ],
        with_watcher: bool = False,
    ) -> Tuple[Diagram, Context]:
        """
        *Description*
        
        This function is used to easily cast and build the production.

        *Parameters*

        planning_algorithm: Callable[[np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],Tuple[MotionPlan, bool]]
            The algorithm that we will use to plan the motion.

        with_watcher: A Boolean that determines whether to add a watcher to the diagram.
            by default False.

        *Returns*

        diagram: Diagram
            The built diagram.

        diagram_context: Context
            The context for the built diagram.
            If with_watcher is True, this context will be used to initialize the watcher.
        """
        # Setup

        # Use Base class implementation to start
        diagram, diagram_context = super().easy_cast_and_build(
            planning_algorithm,
            with_watcher=with_watcher,
        )

        # print("completed cast and build.")

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        # self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
        #     self.station.arm_controller.plant, diagram_context,
        # )

        self.performers[0].set_internal_root_context(
            diagram_context
        )

        # # Set the start configuration of the robot
        # arm = self.station.arm
        # self.station.plant.SetDefaultPositions(
        #     # self.station.plant.GetMyMutableContextFromRoot(diagram_context),
        #     arm,
        #     self.start_configuration,
        # )

        # # Set the positions of the gripper
        # gripper = self.station.gripper
        # n_gripper_positions = self.station.plant.num_positions(gripper)
        # self.station.plant.SetPositions(
        #     self.station.plant.GetMyMutableContextFromRoot(diagram_context),
        #     gripper,
        #     np.zeros((n_gripper_positions,)),
        # )

        return diagram, diagram_context

    @property
    def goal_configuration(self) -> np.ndarray:
        """
        *Description*

        Get the goal pose.
        
        *Returns*

        q_goal: np.ndarray
            The goal configuration of the robot.
        """
        # Setup
        hardcoded_robot_joint_names = [
            'ur10e-test_shoulder_pan_joint_q',
            'ur10e-test_shoulder_lift_joint_q',
            'ur10e-test_elbow_joint_q',
            'ur10e-test_wrist_1_joint_q',
            'ur10e-test_wrist_2_joint_q',
            'ur10e-test_wrist_3_joint_q',
        ]

        # Retrieve goal_configuraiton value
        if self._goal_config is None:
            pose_WorldGoal = self.pose_WorldBeaker.multiply(self.pose_BeakerGoal)

            # Use Inverse Kinematics to get the goal configuration of the robot
            self._goal_config = self.solve_pose_ik_problem(
                pose_WorldGoal,
                robot_joint_names=hardcoded_robot_joint_names,
            )

            return self._goal_config
        else:
            return self._goal_config

    @property
    def goal_pose(self) -> RigidTransform:
        """
        *Description*
        
        Get the goal pose of the end effector frame.

        *Returns*

        pose_WorldGoal: RigidTransform
            The goal pose of the end effector frame.
        """
        # Setup

        # Algorithm
        if self._goal_pose is not None:
            return self._goal_pose
        elif (self._goal_pose is None) and (self._goal_config is None):
            pose_WorldGoal = self.pose_WorldBeaker.multiply(self.pose_BeakerGoal)
            self._goal_pose = pose_WorldGoal

            return self._goal_pose
        elif (self._goal_pose is None) and (self._goal_config is not None):
            # If the goal configuration is given,
            # use the forward kinematics solver to get the goal pose
            self._goal_pose = self.solve_forward_kinematics_problem_for_arm(self._goal_config)
            return self._goal_pose
        else:
            raise ValueError(
                f"It appears that the starting configuration is set, but the starting pose is not. This is unexpected behavior!"
            )

    @property
    def id(self) -> ProductionID:
        """
        *Description*

        Always returns the ProductionID for ChemLab2.

        *Returns*

        ProductionID.kChemLab2: ProductionID
            The ProductionID for ChemLab2.
        """
        return ProductionID.kChemLab2
        
    def initialize_pose_data(self):
        """
        *Description*

        Initialize the following poses if they are not already set:

        - pose_WorldShelf
        - pose_WorldBeaker
        - pose_WorldHolder
        - pose_BeakerGoal
        - pose_HolderStart
        """
        # Set shelf pose
        if self.shelf_pose is None:
            self.pose_WorldShelf = RigidTransform(
                RollPitchYaw(0.0, 0.0, -np.pi/2.0).ToQuaternion(),
                np.array([0.0, 0.75, 0.66]),
            )

        # Set Beaker Pose default
        if self.pose_WorldBeaker is None:
            self.pose_WorldBeaker = RigidTransform(
                RollPitchYaw(np.pi/2.0, 0.0, 0.0).ToQuaternion(),
                np.array([-0.6, 0.45, 0.175]),
            )

        # Set Holder Pose
        self.pose_WorldHolder = RigidTransform(
            RollPitchYaw(np.pi/2.0, 0.0, 0.0).ToQuaternion(),
            np.array([self.table_width*0.5*0.7, 0.6+self.table_length/4., self.table_height+0.015]),
        )

        # Define pose of the goal
        beaker_to_goal_translation = np.array([+0.0, 0.2, 0.0]) 
        beaker_to_goal_orientation = RollPitchYaw(np.pi, np.pi, 0.0).ToQuaternion()
        self.pose_BeakerGoal = RigidTransform(  
            beaker_to_goal_orientation,
            beaker_to_goal_translation,
        )

        # Define pose of start wrt holder
        holder_to_start_translation = np.array([+0.0, 0.2, 0.025])
        holder_to_start_orientation = RollPitchYaw(0.0, 0.0, np.pi).ToQuaternion()
        self.pose_HolderStart = RigidTransform(
            holder_to_start_orientation,
            holder_to_start_translation,
        )

    def solve_forward_kinematics_problem_for_arm(
        self,
        robot_joint_positions: np.ndarray,
    ) -> RigidTransform:
        """
        *Description*
        
        This method solves the forward kinematics problem for the UR10e arm
        by itself.
        """
        # Setup
        target_frame_name = "tool0"

        # Create shadow plant and populate it with the UR10e arm
        shadow_station = UR10eStation(
            time_step=self.time_step,
            meshcat_port_number=None,
            gripper_type=self.gripper_type,
        )
        shadow_station.Finalize()

        # Use station's plant to solve the forward kinematics problem
        shadow_plant = shadow_station.plant
        
        # Set the arm's joint positions
        temp_context = shadow_plant.CreateDefaultContext()
        shadow_plant.SetPositions(
            temp_context,
            shadow_station.arm,
            robot_joint_positions,
        )

        # Get the pose of the end effector
        end_effector_pose = shadow_plant.EvalBodyPoseInWorld(
            temp_context,
            shadow_plant.GetBodyByName(target_frame_name),
        )

        return end_effector_pose

    @property
    def start_configuration(self):
        """
        *Description*

        Get the start configuration of the robot in the production.

        *Returns*

        q_start: np.ndarray
            The start configuration of the robot.
        """
        # Setup
        hardcoded_robot_joint_names = [
            'ur10e-test_shoulder_pan_joint_q',
            'ur10e-test_shoulder_lift_joint_q',
            'ur10e-test_elbow_joint_q',
            'ur10e-test_wrist_1_joint_q',
            'ur10e-test_wrist_2_joint_q',
            'ur10e-test_wrist_3_joint_q',
        ]

        # Algorithm
        if self._start_config is not None:
            return self._start_config
        elif self._start_pose is not None:
            # Use the start pose to get the start configuration
            # Using the IK solver (potentially buggy because default ik problem ignores obstacles)
            return self.solve_pose_ik_problem(
                self._start_pose,
                robot_joint_names=hardcoded_robot_joint_names,
            )
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )


    @property
    def start_pose(self) -> RigidTransform:
        """
        *Description*

        Get the start pose. This should be defined by the subclass.

        *Returns*

        pose_WorldStart: RigidTransform
            The start pose of the end effector frame.
        """
        # Setup

        # Algorithm
        if not (self._start_pose is None):
            return self._start_pose
        elif (self._start_pose is None) and (self._start_config is None):
            # Define Start Pose
            pose_WorldStart = self.pose_WorldHolder.multiply(self.pose_HolderStart)
            self._start_pose = pose_WorldStart
            return self._start_pose
        elif (self._start_pose is None) and (self._start_config is not None):
            # Use the start configuration to get the starting pose
            # Using a "forward kinematics solver"
            self._start_pose = self.solve_forward_kinematics_problem_for_arm(self._start_config)
            return self._start_pose
        else:
            raise ValueError(
                f"Unexpected behavior. This should never happen."
            )



from dataclasses import dataclass
import networkx as nx
from typing import List

# Internal Imports
from brom_drake.utils.leaf_systems.network_fsm import (
    NetworkXFSM, FSMOutputDefinition, FSMTransitionCondition, FSMTransitionConditionType
)
from brom_drake.utils.pick_and_place.phase import PickAndPlacePhase
from brom_drake.utils.pick_and_place.target_description import PickAndPlaceTargetDescription

@dataclass
class Script:
    """
    Description
    -----------
    This class defines the "script" or the sequence of events that is meant to happen in the
    Pick and Place (or Motion Planning and Grasping) task.
    
    Assumptions
    -----------
    We assume that the robot will try to grasp each of the targets in the order that they
    appear in the list below. (i.e., pick up target 0, then target 1, etc.)
    """
    grasping_targets: List[PickAndPlaceTargetDescription]
    time_outs: List[float] # The amount of time given for the robot to reach each of the targets
    initial_settling_time: float # The time after initialization before the robot begins planning or moving
    
    def to_networkx_graph(self) -> nx.DiGraph:
        # Setup
        graph = nx.DiGraph()

        # Create initial state
        graph.add_node("init")

        # Collect the names of all of the targets
        target_names = []
        for target_name_ii in self.grasping_targets:
            name_candidate_ii = target_name_ii.name
            if name_candidate_ii in target_names:
                raise NotImplementedError(
                    f"Unsure of how to add a second object with name \"{name_candidate_ii}\" to the list of targets for the MotionPlanningAndGraspingProductionScript."
                )
            
            target_names.append(name_candidate_ii)

        n_targets = len(target_names)

        # For each target name, we're going to create the following PickAndPlace states.
        pick_and_place_phases = [
            PickAndPlacePhase.kPreGrasp, PickAndPlacePhase.kGrasp, PickAndPlacePhase.kPostGrasp,
            PickAndPlacePhase.kPrePlace, PickAndPlacePhase.kPlace, PickAndPlacePhase.kPostPlace
        ]
        for ii, target_name_ii in enumerate(target_names):
            target_ii: PickAndPlaceTargetDescription = self.grasping_targets[ii]

            # Check the value of the model instance index
            assert target_ii.model_instance_index is not None, \
                f"The provided target does not have a model instance index defined yet. Please define one before calling to_networkx_graph()!"

            for phase_jj in pick_and_place_phases:
                graph.add_node(
                    target_ii.name_for_pick_and_place_phase(phase_jj),
                    outputs=[
                        FSMOutputDefinition(f"model_target", target_ii.model_instance_index),
                    ]
                )

        # Add edges
            
        return graph

    def to_networkx_fsm(self) -> NetworkXFSM:
        pass


from dataclasses import dataclass

@dataclass
class InertiaDefinition:
    """
    *Description*

    A dataclass that defines the inertia of a simple shape.

    This is defined to match the URDF specifications for inertia.

    *Attributes*

    ixx: float
        The xx component of the inertia tensor.

    ixy: float
        The xy component of the inertia tensor.

    ixz: float
        The xz component of the inertia tensor.

    iyy: float
        The yy component of the inertia tensor.

    iyz: float
        The yz component of the inertia tensor.

    izz: float
        The zz component of the inertia tensor.
    """
    ixx: float = 1.0
    ixy: float = 0.0
    ixz: float = 0.0
    iyy: float = 1.0
    iyz: float = 0.0
    izz: float = 1.0

    def as_list(self) -> list[float]:
        """
        *Description*

        Convert the inertia to a list.

        *Returns*

        list[float]
            A list of the inertia values.
        """
        return [self.ixx, self.ixy, self.ixz, self.iyy, self.iyz, self.izz]

    def as_map(self) -> dict[str, str]:
        """
        *Description*

        Convert the inertia to a map.

        *Returns*

        dict[str, str]
            A map of the inertia values.
        """
        return {
            "ixx": f"{self.ixx}",
            "ixy": f"{self.ixy}",
            "ixz": f"{self.ixz}",
            "iyy": f"{self.iyy}",
            "iyz": f"{self.iyz}",
            "izz": f"{self.izz}",
        }


from dataclasses import dataclass
from enum import IntEnum
import numpy as np
from typing import Union
import xml.etree.ElementTree as ET


class ShapeEnum(IntEnum):
    """
    *Description*
    
    The different simple shapes that can be created.
    """
    kBox = 1
    kSphere = 2
    kCylinder = 3
    kCapsule = 4

# Shape Definition Can be any one of these
@dataclass
class ShapeDefinition:
    """
    *Description*
    
    Base class for defining a shape as per URDF specifications.
    """
    def add_geometry_to_element(self, target_element: ET.Element) -> ET.Element:
        """
        *Description*

        Adds the shape geometry as a sub-element to the given target element.

        This method must be implemented in subclasses.

        *Parameters*

        target_element: xml.etree.ElementTree.Element
            The target XML element to which the shape geometry will be added.

        *Returns*

        xml.etree.ElementTree.Element
            The element `target_element` with the shape geometry added as a sub-element.
        """
        raise NotImplementedError("This method must be implemented in the subclass.")

    @property
    def type(self) -> ShapeEnum:
        """
        *Description*

        Returns the ShapeEnum type corresponding to the shape.

        This method must be implemented in subclasses.

        *Returns*

        ShapeEnum
            The ShapeEnum type for the shape.
        """
        raise NotImplementedError("This method must be implemented in the subclass.")



from brom_drake.productions.types.pick_and_place.mp_and_grasping import (
    MotionPlanningAndGraspingProduction,
)
from .script import Script as MotionPlanningAndGraspingProductionScript

__all__ = [
    'MotionPlanningAndGraspingProduction',
    'MotionPlanningAndGraspingProductionScript',
]


"""
motion_planner.py
Description:

        This file defines the MotionPlanner class.
        This class is used to plan motion for a robot.
"""
import numpy as np
from pydrake.geometry import SceneGraph
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import ModelInstanceIndex


class MotionPlanner:
    def __init__(
        self,
        robot_model_idx: ModelInstanceIndex,
        plant: MultibodyPlant,
        scene_graph: SceneGraph,
        random_seed: int = 23,
    ):
        """
        Description:
            This function initializes the motion planner.
        """
        # Input Processing
        self.robot_model_idx = robot_model_idx
        self.plant = plant
        self.scene_graph = scene_graph

        # Setup
        self.root_context = None # Usually, the diagram context that we can use to extract the plant's context

        # Set the random seed
        self.random_seed = random_seed
        np.random.seed(self.random_seed)

    @property
    def dim_q(self) -> int:
        if not self.plant.is_finalized():
            raise ValueError("Plant has not been finalized yet! Can not compute num_actuated_dofs().")

        return self.plant.num_actuated_dofs(self.robot_model_idx)

    def check_collision_in_config(
        self,
        q_model: np.ndarray,
    ) -> bool:
        """
        Description:
            This function checks for collisions in the robot's environment.
        """
        # Input Processing
        if self.root_context is None:
            raise ValueError("Plant context is not initialized yet!")

        # Setup
        plant_context = self.plant.GetMyMutableContextFromRoot(self.root_context)
        scene_graph_context = self.scene_graph.GetMyMutableContextFromRoot(self.root_context)

        # Set the configuration
        self.plant.SetPositions(
            plant_context,
            self.robot_model_idx,
            q_model
        )

        # Note: This method will be inaccurate if using arbitrary, nonconvex meshes!
        # https://drake.mit.edu/pydrake/pydrake.geometry.html?highlight=queryobject#pydrake.geometry.QueryObject.HasCollisions
        query_object = self.scene_graph.get_query_output_port().Eval(scene_graph_context)
        # return query_object.HasCollisions()

        # Alternative method to check for collisions
        closest_points = query_object.ComputeSignedDistancePairwiseClosestPoints()
        eps0 = 1e-2
        for pair in closest_points:
            if pair.distance < eps0:
                return True
            
        # Otherwise return false
        return False
    
    @property
    def joint_limits(self) -> np.ndarray:
        """
        Description:
            This function retrieves the joint limits of the robot.
        """
        # Setup
        joint_indicies = self.plant.GetJointIndices(self.robot_model_idx)

        joint_limits = np.zeros((0, 2))
        for ii, joint_index in enumerate(joint_indicies):
            # Check to see if joint is actuated; if not, then ignore
            joint_ii = self.plant.get_joint(joint_index)
            joint_can_move = joint_ii.can_rotate() or joint_ii.can_translate()
            if not joint_can_move:
                continue

            lower_limits = self.plant.get_joint(joint_index).position_lower_limits()
            if len(lower_limits) == 0:
                lower_limits = [-np.inf]

            upper_limits = self.plant.get_joint(joint_index).position_upper_limits()
            if len(upper_limits) == 0:
                upper_limits = [np.inf]

            # Append limits to the joint_limits array
            joint_limits = np.vstack((
                joint_limits,
                [lower_limits[0], upper_limits[0]]
            ))

        return joint_limits

    def sample_random_configuration(self) -> np.ndarray:
        """
        Description
        -----------
        This function samples a random configuration within the joint limits.
        """
        return np.random.uniform(
            self.joint_limits[:, 0],
            self.joint_limits[:, 1]
        )



from pydrake.all import (
    MultibodyPlant,
    ModelInstanceIndex,
    Parser,
    RigidBody,
)

def find_number_of_positions_in_welded_model(
    path_to_model: str,
) -> int:
    """
    Description
    -----------
    This method will return the number of positions in the 
    model described at path_to_model WHEN the base link is welded
    to the world frame.

    This is most useful for finding things like the number of positions in
    a robot.

    Returns
    -------
    int
        The number of positions in the model.
    """
    # Setup
    time_step = 1e-3

    # Create a shadow plant
    shadow_plant = MultibodyPlant(time_step)
    model_idcs = Parser(plant=shadow_plant).AddModels(path_to_model)
    shadow_model_idx = model_idcs[0]

    # Weld the base link to the world frame
    gripper_bodies_indicies = shadow_plant.GetBodyIndices(shadow_model_idx)
    first_body_in_gripper = shadow_plant.get_body(gripper_bodies_indicies[0])
    shadow_plant.WeldFrames(
        shadow_plant.world_frame(),
        first_body_in_gripper.body_frame(),
    )

    # Finalize the shadow plant
    shadow_plant.Finalize()

    return shadow_plant.num_positions()

def get_first_body_in(plant: MultibodyPlant, model_instance: ModelInstanceIndex) -> RigidBody:
    """
    *Description*
    
    This method returns the first body in the given model instance.
    """
    return get_all_bodies_in(plant, model_instance)[0]

def get_all_bodies_in(
    plant: MultibodyPlant,
    model_instance: ModelInstanceIndex,
) -> list[RigidBody]:
    """
    *Description*
    
    This method returns all the bodies in the given model instance.
    """
    # Setup

    # Algorithm
    model_body_indices = plant.GetBodyIndices(model_instance)
    return [plant.get_body(body_index) for body_index in model_body_indices]

def get_name_of_first_body_in_urdf(urdf_in: str) -> str:
    """
    Description
    -----------
    This method returns the name of the first body found in the 
    urdf given by urdf_in.
    """
    return get_name_of_all_bodies_in_urdf(urdf_in)[0]

def get_name_of_all_bodies_in_urdf(urdf_in: str) -> list[str]:
    """
    *Description*
    
    This method returns the name of the first body found in the 
    urdf given by urdf_in.
    """
    # Setup

    # Create Plant with this urdf
    plant = MultibodyPlant(time_step=1e-3)
    try:
        model_instances = Parser(plant).AddModels(urdf_in)
    except Exception as e:
        raise RuntimeError(
            f"Failed to parse the urdf file: {urdf_in}."
        ) from e

    if len(model_instances) > 1:
        raise RuntimeWarning(
            f"More than one model instance found in the urdf file: {urdf_in}."
        )
    elif len(model_instances) == 0:
        raise RuntimeWarning(
            f"No model instance found in the urdf file: {urdf_in}."
        )

    # Algorithm
    model_instance = model_instances[0] # TODO(kwesi): Do we need to finalize the plant?
    all_bodies = get_all_bodies_in(plant, model_instance)
    return [body.name() for body in all_bodies]


from pathlib import Path
from pydrake.all import Quaternion, RigidTransform
import numpy as np
from typing import Union
from xml.etree import ElementTree as ET

# Internal imports
from brom_drake.file_manipulation.urdf.shapes.shape_definition import ShapeDefinition

class URDFElementCreator:
    """
    A class to create URDF elements quickly and easily.
    """
    @staticmethod
    def create_collision_element(
        name: str,
        pose_ParentMesh: Union[None, np.ndarray, RigidTransform] = None,
        mesh_file_path: Union[str, None] = None,
        mesh_scale: np.ndarray = np.array([1.0, 1.0, 1.0]),
    ) -> ET.Element:
        """
        Create a collision element for the URDF.
        
        Parameters
        ----------
        name: str
            The name of the collision element.
        pose_ParentMesh: Union[None, np.ndarray, RigidTransform], optional
            The pose of the parent mesh in the collision element. If None, a default RigidTransform is used.
            If provided as a numpy array, it should be a 7-element array representing position and quaternion.
            (Specifically, the first 3 elements are the position, and the last 4 elements are the quaternion [x,y,z,w].)
        
        Returns
        -------
        ET.Element
            The created collision element.
        """
        # Setup

        if pose_ParentMesh is None:
            pose_ParentMesh = RigidTransform()
        
        if isinstance(pose_ParentMesh, np.ndarray):
            assert pose_ParentMesh.shape == (7,), \
                "pose_ParentMesh should be a 7-element array representing a RigidTransform."
            pose_ParentMesh: RigidTransform = RigidTransform(
                p = pose_ParentMesh[:3],  # Position (x, y, z)
                quaternion= Quaternion(pose_ParentMesh[[-1,3,4,5]]) # Quaternion given as (w, x, y, z)
            )

        if mesh_file_path is None:
            raise NotImplementedError(
                f"Mesh file path is currently required for creating a collision element."
            )

        # Create the collision element
        collision = ET.Element("collision", {"name": name})

        # Add the origin sub-element to the collision element
        collision.append(
            URDFElementCreator.create_origin_element(
                pose_ParentOrigin=pose_ParentMesh
            )
        )

        # If a mesh file path is provided, add a geometry sub-element
        # with a mesh sub-element inside it
        if mesh_file_path is not None:
            collision.append(
                URDFElementCreator.create_geometry_element(
                    mesh_file_path=mesh_file_path,
                    mesh_scale=mesh_scale,
                )
            )

        return collision
    
    @staticmethod
    def create_origin_element(
        pose_ParentOrigin: RigidTransform,
    ) -> ET.Element:
        """
        Create an origin element for the URDF.
        
        Parameters
        ----------
        translation: np.ndarray
            The translation vector (x, y, z).
        rotation: Quaternion
            The rotation as a Quaternion.
        
        Returns
        -------
        ET.Element
            The created origin element.
        """
        # Setup
        translation = pose_ParentOrigin.translation()
        rotation = pose_ParentOrigin.rotation()
        rotation_as_rpy = rotation.ToRollPitchYaw()

        # Create the origin element
        origin = ET.Element("origin")
        origin.set("xyz", f"{translation[0]} {translation[1]} {translation[2]}")
        origin.set("rpy", f"{rotation_as_rpy.roll_angle()} {rotation_as_rpy.pitch_angle()} {rotation_as_rpy.yaw_angle()}")
        
        return origin
    
    @staticmethod   
    def create_geometry_element(
        mesh_file_path: str = None,
        definition: ShapeDefinition = None,
        mesh_scale: np.ndarray = np.array([1.0, 1.0, 1.0])
    ) -> ET.Element:
        """
        Create a geometry element for the URDF.
        
        Parameters
        ----------
        mesh_file_path: Path, optional
            The path to the mesh file. If provided, a mesh sub-element will be created.
        definition: ShapeDefinition, optional
            The dimensions of the geometry. Required for box and cylinder types.
        
        Returns
        -------
        ET.Element
            The created geometry element.
        """
        # Create the geometry element
        geometry = ET.Element("geometry")
        
        # If a path to a mesh file is provided, add a mesh sub-element
        if mesh_file_path is not None:
            mesh = ET.SubElement(geometry, "mesh")
            mesh.set("filename", str(mesh_file_path))
            # # If the mesh file path is relative, make it absolute
            # if not str(mesh_file_path).startswith("/"):
            #     mesh.set("filename", f"{mesh_file_path}")

            # Set the scale for the mesh
            mesh.set("scale", f"{mesh_scale[0]} {mesh_scale[1]} {mesh_scale[2]}")
            

        # If a shape definition is provided, add the appropriate sub-element
        # TODO: Implement support for different shape definitions
        
        return geometry



from .bool_to_vec_system import BoolToVectorSystem
from .end_effector_wrench_calculator import EndEffectorWrenchCalculator
from .named_vector_selection_system import define_named_vector_selection_system
from .rigid_transform_to_vector_system.system import RigidTransformToVectorSystem
from brom_drake.utils.leaf_systems.flexible_port_switch import FlexiblePortSwitch
from brom_drake.utils.leaf_systems.network_fsm import (
    FSMEdgeDefinition,
    FSMOutputDefinition,
    FSMTransitionCondition,
    FSMTransitionConditionType,
    NetworkXFSM,
)

__all__ = [
    "BoolToVectorSystem",
    "define_named_vector_selection_system",
    "EndEffectorWrenchCalculator",
    "FlexiblePortSwitch",
    "FSMEdgeDefinition",
    "FSMOutputDefinition",
    "FSMTransitionCondition",
    "FSMTransitionConditionType",
    "NetworkXFSM",
    "RigidTransformToVectorSystem",
]


from .config import ProximityPosePlanDispenserConfig
from .dispenser_internal_state import DispenserInternalState, DispenserTransitionRequest
from .proximity_pose_plan_dispenser import ProximityPosePlanDispenser

__all__ = [
    "ProximityPosePlanDispenserConfig",
    "DispenserInternalState",
    "DispenserTransitionRequest",
    "ProximityPosePlanDispenser",
]


"""
Description:
    This file contains a set of helpful utilities that can be used to help with the
    development of the robot models.
"""

from enum import IntEnum
from pathlib import Path
from typing import List
import xml.etree.ElementTree as ET

class BaseLinkSearchApproach(IntEnum):
    kIncludesBaseInName = 1
    kFirstLinkFound = 2

def find_base_link_name_in(path_to_robot_model: str|Path, search_method: BaseLinkSearchApproach = BaseLinkSearchApproach.kFirstLinkFound) -> str:
    """
    *Description*
    
    This method will try to find a good base link for the model.
    
    *Parameters*

    path_to_robot_model: str|Path
        The path to the robot model.
    
    *Returns*

    base_link_name: str
        The name of the link that we believe is the base.
    """
    # Setup
    if type(path_to_robot_model) is str:
        path_to_robot_model = Path(path_to_robot_model)

    path_to_model = path_to_robot_model

    # Parse the model using xml
    if (".urdf" in path_to_model.name) or (".sdf" in path_to_model.name):
        original_xml = ET.ElementTree(file=str(path_to_model))
        link_names = find_all_link_names(original_xml)

        # Find a link that contains the word "base"
        match search_method:
            case BaseLinkSearchApproach.kIncludesBaseInName:
                for link_name in link_names:
                    if "base" in link_name.lower():
                        return link_name # Return the first one we find
                    
                
            case BaseLinkSearchApproach.kFirstLinkFound:
                return link_names[0]

    else:
        raise ValueError(
            "We can only smartly find base links in .urdf files, for now.\n" +
            "File an issue if you want more support in the future."
        )

    # If we can't find a base link, then we will raise an error
    raise ValueError(
        "We could not find a good base link in the model.\n" +
        "Please provide the base link name manually by adding \"base\" to one of the urdf's links."
    )


def find_all_link_names(xml_tree: ET.ElementTree) -> List[str]:
    """
    **Description**
    
    This method will find all the link names in the xml tree ``xml_tree``.

    **Parameters**
    
    xml_tree: xml.etree.ElementTree.ElementTree
        The xml tree that we would like to investigate.

    **Returns**

    link_names: List[str]
        A list of all the names of <link> tags in the model.
    """
    # Setup
    link_names = []

    # Find all the links
    for link in xml_tree.findall(".//link"):
        link_names.append(link.attrib["name"])

    return link_names


from dataclasses import dataclass
import numpy as np

@dataclass
class PlanningNode:
    id: int
    q: np.ndarray

    def __hash__(self):
        return hash(self.id)
    
    def __eq__(self, other):
        return self.id == other.id


import numpy as np
from pydrake.all import AffineSystem
from typing import List

def define_named_vector_selection_system(
    all_element_names: List[str],
    sequence_of_names_for_output: List[str],
) -> AffineSystem:
    """
    *Description*
    
    This method creates an Affine System that will receive as input
    a vector where each element is named according to `all_element_names`
    and returns an output vector that selects only the elements
    specified in `desired_inputs`.

    *Parameters*
    
    name : str
        The name of the system to be added.

    all_element_names : List[str]
        The list of all element names in the input vector.

    desired_inputs : List[str]
        The list of desired element names to be selected for output.

    *Returns*
    
    AffineSystem
        The created NamedVectorSelectionSystem.
    """
    # Setup
    n_x = len(all_element_names)
    n_y = len(sequence_of_names_for_output)

    # Create the selection matrix
    D = np.zeros((n_y, n_x))
    for i, desired_name in enumerate(sequence_of_names_for_output):
        if desired_name not in all_element_names:
            raise ValueError(
                f"Desired input '{desired_name}' not found in all_element_names."
            )
        j = all_element_names.index(desired_name)
        D[i, j] = 1.0

    # Create the Affine System
    return AffineSystem(
        D=D,
    )


from pydrake.all import RigidTransform
from typing import Any

def is_rigid_transform(obj: Any) -> bool:
    """
    Description
    -----------
    This function checks if the object is a RigidTransform.
    We will perform 2 checks.
    1. Check if the object is a RigidTransform object directly, and then
    2. Check if the object has the translation() and rotation() methods.
    """
    # Check if the object is a RigidTransform object directly
    if isinstance(obj, RigidTransform):
        return True
    
    # Check if the object has the translation() and rotation() methods
    if hasattr(obj, "translation") and hasattr(obj, "rotation"):
        # Check to see if the translation and rotation methods are callable
        if callable(obj.translation) and callable(obj.rotation):
            return True
        
    # Otherwise, return False
    return False



import numpy as np
from pydrake.all import (
    AbstractValue, BasicVector,
    Context,
    LeafSystem,
)

class BoolToVectorSystem(LeafSystem):
    """
    *Description*
    
    This LeafSystem receives boolean inputs (i.e., AbstractValue inputs
    containing bool values) and returns an output vector of size 1 (i.e., BasicVector) where the
    single element is 1 if the input bool is True and 0 if the input bool
    is False.

    *Diagram*
    
    The LeafSystem's input and output ports can be illustrated with the following block: ::

                        |---------------|
                        | Bool          |
        bool_in ---->   | To            | ---> vector_out 
        (bool)          | VectorSystem  | (BasicVector[1])
                        |---------------|

    """
    def __init__(self):
        LeafSystem.__init__(self)
        self.boolean_input_port = self.DeclareAbstractInputPort(
            "bool_in",
            AbstractValue.Make(False),
        )
        self.DeclareVectorOutputPort(
            "vector_xyz_quat(wxyz)",
            1,
            self.CalcVectorOutput,
        )

    def CalcVectorOutput(self, context: Context, output: BasicVector):
        """
        *Description*
        
        Callback function for calculating the output vector of the `BoolToVectorSystem`.
        This function takes in a bool and outputs a vector of size 1.
        """
        # Setup

        # Get the bool input
        bool_input = self.boolean_input_port.Eval(context)
        output.SetFromVector(
            np.array([int(bool_input)])
        )


from dataclasses import dataclass
from enum import IntEnum
import networkx as nx
import numpy as np
from typing import List, Union

# Internal Imports
from brom_drake.utils.leaf_systems.network_fsm.fsm_transition_condition import FSMTransitionCondition

class FSMEdgeDefinition:
    """
    Description
    """
    def __init__(
        self,
        conditions: List[FSMTransitionCondition],
        src: int,
        dst: int,
    ):
        # Setup

        # Input Checking
        assert len(conditions) > 0, "At least one condition must be provided."

        # Assign Attributes
        self.conditions = conditions
        self.src = src
        self.dst = dst





from importlib import resources as impresources
import networkx as nx
import numpy as np
from pydrake.all import (
    AbstractValue,
    CollisionFilterDeclaration,
    ConstantValueSource,
    ConstantVectorSource,
    Context,
    Diagram,
    GeometrySet,
    InverseKinematics,
    MultibodyPlant,
    Parser,
    Quaternion,
    RigidTransform,
    RotationMatrix,
)
from pydrake.math import RollPitchYaw, RigidTransform
from typing import Callable, Tuple

# Internal Imports
from brom_drake.directories import DEFAULT_BROM_MODELS_DIR
from brom_drake.control import (
    GripperTarget,
)
from brom_drake.file_manipulation.urdf import drakeify_my_urdf
from brom_drake.file_manipulation.urdf.shapes.box import BoxDefinition
from brom_drake.file_manipulation.urdf.simple_writer.urdf_definition import SimpleShapeURDFDefinition, \
    InertiaDefinition
from brom_drake.motion_planning.systems.open_loop_dispensers.open_loop_plan_dispenser import OpenLoopPlanDispenser
import brom_drake.robots as robots
from brom_drake.robots.gripper_type import GripperType
from brom_drake.stations.classical import UR10eStation
from brom_drake.productions import ProductionID
from brom_drake.productions.roles import Role
from brom_drake.productions.types import MotionPlanningAndGraspingProduction
from brom_drake.utils import Performer, AddGround, MotionPlan


class ChemLab3(MotionPlanningAndGraspingProduction):
    """
    Description:
        This production is the first in the chemistry lab series.
        It is used to test the motion planning capabilities of the robot
        in a chemistry lab setting with minimal constraints.
    """
    def __init__(
        self,
        time_step=5e-4,
        meshcat_port_number: int = 7001,
        plan_execution_speed: float = 0.2,
        shelf_pose: RigidTransform = None,
        gripper_type: GripperType = GripperType.Robotiq_2f_85,
        table_length: float = 0.9,
        table_width: float = 2.0,
        table_height: float = 0.1,
        **kwargs,
    ):
        """
        Description:
            Constructor for the ChemLab3 Production.
        :param meshcat_port_number:
        """
        # Superclass constructor
        super().__init__(**kwargs)

        # Input Processing
        self.time_step = time_step
        self.meshcat_port_number = meshcat_port_number
        self.plan_execution_speed = plan_execution_speed
        self.table_length = table_length
        self.table_width = table_width
        self.table_height = table_height
        self.gripper_type = gripper_type

        # Initialize pose data
        self.shelf_pose = shelf_pose
        self.pose_WorldFlask0 = None
        self.pose_WorldTableHalf1, self.pose_WorldTableHalf2 = None, None
        self.initialize_pose_data()

        # Set station
        self.station = UR10eStation(
            time_step=self.time_step,
            meshcat_port_number=self.meshcat_port_number,
            gripper_type=self.gripper_type,
        )
        self.arm = self.station.arm
        self.robot_model_idx_ = self.arm

        # self.gripper = self.station.gripper

        # Set Names of Plant and scene graph
        self.plant = self.station.plant
        self.scene_graph = self.station.scene_graph

        self.plant.set_name(f"ChemLab3_Production_Plant")
        self.scene_graph.set_name(f"ChemLab3_Production_SceneGraph")

        # Define placeholder variables for models
        self.test_tube_holder1 = None
        self.flask_model_index = None

    def add_supporting_cast(self):
        """
        Description
        -----------
        This method adds all secondary cast members to the builder.
        The secondary cast members in the production are the:
        - Table, where the robot exists
        - Test Tube Holders
        - Component which share's the robot model reference
        - Motion Planning components (e.g., dispensers, etc.)
        - Start and Goal sources
        :return:
        """
        # Call the superclass method
        super().add_supporting_cast()

        # Setup

        # Create the table and the items on it
        self.add_table()
        self.add_erlenmeyer_flask()
        self.add_shelf()
        AddGround(self.plant)

        # Add the station
        self.builder.AddSystem(
            self.station,
        )

        self.station.Finalize()

        # Add the motion planning components
        self.add_robot_source_system()
        self.add_motion_planning_components()
        self.add_start_and_goal_sources_to_builder()
        # self.add_dummy_gripper_components()

        print("Completed adding supporting cast members.")

    def add_dummy_gripper_components(self):
        """
        Description
        -----------
        Add more components to the production that are used
        to hold the gripper in place.
        """
        # Setup

        # Create a dummy target type object for the gripper and connect it to the gripper controller
        dummy_gripper_target_type_source = self.builder.AddSystem(
            ConstantValueSource(
                AbstractValue.Make(GripperTarget.kPosition)
            ),
        )
        
        self.builder.Connect(
            dummy_gripper_target_type_source.get_output_port(),
            self.station.GetInputPort("gripper_target_type"),
        )

        # Create a dummy gripper target value and connect it to the gripper controller
        dummy_gripper_target_source = self.builder.AddSystem(
            ConstantVectorSource(np.array([0.0])),
        )
        self.builder.Connect(
            dummy_gripper_target_source.get_output_port(),
            self.station.GetInputPort("gripper_target"),
        )

    def add_erlenmeyer_flask(self):
        """
        Description
        -----------
        Adds the 500ml Erlenmeyer flask to the production.
        """
        # Setup
        plant = self.plant
        urdf_file_path = str(
            impresources.files(robots) / "models/erlenmeyer_flask/500ml.urdf"
        )

        # Use Drakeify my urdf to create the erlenmeyer flask
        new_flask_urdf = drakeify_my_urdf(urdf_file_path)

        # Add the flask to the plant
        model_idcs = Parser(plant=plant).AddModels(str(new_flask_urdf))
        self.flask_model_index = model_idcs[0]

        # Add object to the list we use for initialization
        self.models_in_supporting_cast.append(
            (self.flask_model_index, self.pose_WorldFlask0),
        )

    def add_motion_planning_components(self):
        """
        Add the motion planning components to the builder.
        :return:
        """
        # Setup
        arm = self.arm
        n_actuated_dof = self.plant.num_actuated_dofs(arm) # Number of actuated DOF in arm

        # Add the Plan Dispenser and connect it to the station
        self.plan_dispenser = self.builder.AddSystem(
            OpenLoopPlanDispenser(n_actuated_dof, self.plan_execution_speed)
        )

        self.builder.Connect(
            self.plan_dispenser.GetOutputPort("point_in_plan"),
            self.station.GetInputPort("desired_joint_positions"),
        )

        if self.gripper_type != GripperType.NoGripper:
            # Add components for the gripper's control, including:
            # - Gripper Target Type
            # - Gripper Target
            # Both should be static.
            
            # Add the gripper target to the builder
            gripper_target_type_source = self.builder.AddSystem(
                ConstantValueSource(
                    AbstractValue.Make(GripperTarget.kPosition),
                ),
            )

            # Connect the gripper target type to the station
            self.builder.Connect(
                gripper_target_type_source.get_output_port(),
                self.station.GetInputPort("gripper_target_type"),
            )

            # Add the gripper target to the builder
            gripper_target_source = self.builder.AddSystem(
                ConstantVectorSource(np.array([0.0])),
            )

            # Connect the gripper target to the station
            self.builder.Connect(
                gripper_target_source.get_output_port(),
                self.station.GetInputPort("gripper_target"),
            )


    def add_shelf(self):
        """
        Add the shelf to the production.
        :return:
        """
        # Setup
        plant: MultibodyPlant = self.plant
        urdf_file_path = str(
            impresources.files(robots) / "models/cupboard/cupboard.sdf"
        )

        # Add the shelf to the plant
        model_idcs = Parser(plant=plant).AddModels(urdf_file_path)
        self.shelf_model_index = model_idcs[0]

        # Weld the bookshelf to the world frame
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("cupboard_body", self.shelf_model_index),
            self.pose_WorldShelf,
        )

    def add_table(self):
        """
        Description
        -----------
        This method adds the table to the production.
        The table will be "L-shaped", and we will define it with 2 box
        shapes.
        :return:
        """
        # Setup
        table_width, table_length, table_height = self.table_width, self.table_length, self.table_height

        # 11111111111111111
        # Create first half

        # Create a new shape urdf
        half1_defn = SimpleShapeURDFDefinition(
            name="table_half1",
            shape=BoxDefinition(
                size=(table_width, table_length, table_height),
            ),
            mass=100.0, # kg
            inertia=InertiaDefinition(
                ixx=10.0,
                iyy=10.0,
                izz=10.0,
            ),
            color=np.array([0.1, 0.1, 0.1, 1.0]),
        )
        table_urdf_path = DEFAULT_BROM_MODELS_DIR + "/table/table_half1.urdf"
        half1_defn.write_to_file(table_urdf_path)

        # Add the table to the production
        table_half1_model_index = Parser(self.plant).AddModels(table_urdf_path)[0]

        # Weld the table to the world frame
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName(half1_defn.base_link_name, table_half1_model_index),
            self.pose_WorldTableHalf1,
        )

        # 2222222222222222222222
        # Create the second half

        # Create a new shape urdf
        half2_defn = SimpleShapeURDFDefinition(
            name="table_half2",
            shape=BoxDefinition(
                size=(table_width/3.0, table_length, table_height),
            ),
            mass=50.0, # kg
            inertia=InertiaDefinition(
                ixx=10.0,
                iyy=10.0,
                izz=10.0,
            ),
            color=np.array([0.1, 0.1, 0.1, 1.0]),
        )
        table_urdf_path = DEFAULT_BROM_MODELS_DIR + "/table/table_half2.urdf"
        half2_defn.write_to_file(table_urdf_path)

        # Add the table to the production
        table_half2_model_index = Parser(self.plant).AddModels(table_urdf_path)[0]

        # Weld the table to the world frame
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName(half2_defn.base_link_name, table_half2_model_index),
            self.pose_WorldTableHalf2,
        )

    def add_cast_and_build(
        self,
        cast: Tuple[Role, Performer] = [],
        with_watcher: bool = False,
    ) -> Tuple[Diagram, Context]:
        """
        Description
        -----------
        Modifies the normal add_cast_and_build, so that
        we share the context of the plant with the appropriate
        parts of the system.
        :param cast:
        :param with_watcher: A Boolean that determines whether to add a watcher to the diagram.
        :return:
        """
        diagram, diagram_context = super().add_cast_and_build(
            main_cast_members=cast,
            with_watcher=with_watcher,
        )

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
            self.station.arm_controller.plant, diagram_context,
        )

        for role_ii, performer_ii in cast:
            if role_ii.name == "OfflineMotionPlanner":
                performer_ii.set_internal_root_context(
                    diagram_context
                )

        return diagram, diagram_context
    
    def build_production(
        self,
        with_watcher: bool = True,
    ) -> Tuple[Diagram, Context]:
        """
        Description
        -----------
        This method builds the production.
        It assumes that all components have been added to the builder.

        Arguments
        ---------
        with_watcher: bool
            A Boolean that determines whether to add a watcher to the diagram.
        """
        # Setup

        # Call the parent method
        diagram, diagram_context = super().build_production(with_watcher=with_watcher)

        # Set the initial positions of the shelf
        plant: MultibodyPlant = self.plant
        plant.SetPositions(
            plant.GetMyMutableContextFromRoot(diagram_context),
            self.shelf_model_index,
            np.array([-np.pi*(3.0/4.0), np.pi*(3.0/4.0)]),
        )

        return diagram, diagram_context

    def configure_collision_filter(self, scene_graph_context: Context):
        """
        Description
        -----------
        This method configures the collision filter for the production.
        :param scene_graph_context:
        :return:
        """
        # Setup
        scene_graph = self.scene_graph
        shelf_model_index = self.shelf_model_index

        # Ignore self collisions of the shelf, by:
        # - Getting the model instance index of the shelf
        # - Collecting the Geometry IDs of all the geometries in the shelf
        shelf_geometry_ids = []
        for body_index in self.plant.GetBodyIndices(shelf_model_index):
            # Get the geometry IDs
            shelf_geometry_ids.extend(
                self.plant.GetCollisionGeometriesForBody(
                    self.plant.get_body(body_index)
                )
            )
            # print(self.plant.get_body(body_index))

        self.shelf_geometry_ids = shelf_geometry_ids

        # Apply the collision filter
        scene_graph.collision_filter_manager(scene_graph_context).Apply(
            CollisionFilterDeclaration().ExcludeWithin(
                GeometrySet(self.shelf_geometry_ids)
            )
        )

        # # Ignore collisions between the robot links
        # # TODO: Actually implement this for adjacent links? Or is this not necessary?
        # arm_geometry_ids = []
        # for body_index in self.plant.GetBodyIndices(self.arm):
        #     arm_geometry_ids.extend(
        #         self.plant.GetCollisionGeometriesForBody(
        #             self.plant.get_body(body_index)
        #         )
        #     )

        # self.arm_geometry_ids = arm_geometry_ids

        # # Apply the collision filter
        # scene_graph.collision_filter_manager(scene_graph_context).Apply(
        #     CollisionFilterDeclaration().ExcludeWithin(
        #         GeometrySet(self.arm_geometry_ids)
        #     )
        # )
    
    def define_pose_ik_problem(
        self,
        pose_WorldTarget: RigidTransform,
        target_frame_name: str,
        eps0: float = 2.5e-2,
        orientation_cost_scaling: float = 0.25,
    ) -> InverseKinematics:
        """
        Description
        -----------
        Sets up the inverse kinematics problem for the start pose
        input to theis function.
        :return:
        """
        # Setup
        plant: MultibodyPlant = self.plant
        scene_graph = self.scene_graph
        sg_inspector = scene_graph.model_inspector()

        # Create IK Problem
        ik_problem = InverseKinematics(plant)

        # Add Pose Target
        ik_problem.AddPositionConstraint(
            self.plant.world_frame(),
            pose_WorldTarget.translation(),
            self.plant.GetFrameByName(target_frame_name),
            (- np.ones((3,)) * eps0).reshape((-1, 1)),
            (+ np.ones((3,)) * eps0).reshape((-1, 1)),
        )

        # TODO(kwesi): Add OrientationCosntraint
        # ik_problem.AddOrientationConstraint(
        #     self.plant.world_frame(),
        #     RotationMatrix(Quaternion(input_pose_vec[3:]).rotation()),
        #     self.plant.GetFrameByName("ft_frame"),
        #     RotationMatrix.Identity(),
        #     0.25,
        # )

        # ik_problem.AddOrientationConstraint(
        #     self.plant.world_frame(),
        #     pose_WorldTarget.rotation(),
        #     self.plant.GetFrameByName(target_frame_name),
        #     RotationMatrix.Identity(),
        #     np.pi/8.0
        # )

        ik_problem.AddOrientationCost(
            self.plant.world_frame(),
            pose_WorldTarget.rotation(),
            self.plant.GetFrameByName(target_frame_name),
            RotationMatrix.Identity(),
            orientation_cost_scaling,
        )

        # Create a cost function for the IK problem that weights the joint positions
        # ik_problem.AddMinimumDistanceLowerBoundConstraint(0.05)

        return ik_problem

    def easy_cast_and_build(
        self,
        planning_algorithm: Callable[
            [np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],
            Tuple[MotionPlan, bool],
        ],
        with_watcher: bool = False,
    ) -> Tuple[Diagram, Context]:
        """
        Description
        -----------
        This function is used to easily cast and build the production.
        :param planning_algorithm: The algorithm that we will use to
        plan the motion.
        :param with_watcher: A Boolean that determines whether to add a watcher to the diagram.
        :return:
        """
        # Setup

        # Use Base class implementation to start
        diagram, diagram_context = super().easy_cast_and_build(
            planning_algorithm,
            with_watcher=with_watcher,
        )

        # print("completed cast and build.")

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        # self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
        #     self.station.arm_controller.plant, diagram_context,
        # )

        self.performers[0].set_internal_root_context(
            diagram_context
        )

        return diagram, diagram_context

    @property
    def goal_configuration(self):
        """
        Get the goal pose. This should be defined by the subclass.
        :return:
        """
        # Setup
        hardcoded_robot_joint_names = [
            'ur10e-test_shoulder_pan_joint_q',
            'ur10e-test_shoulder_lift_joint_q',
            'ur10e-test_elbow_joint_q',
            'ur10e-test_wrist_1_joint_q',
            'ur10e-test_wrist_2_joint_q',
            'ur10e-test_wrist_3_joint_q',
        ]

        # Retrieve goal_configuraiton value
        if self._goal_config is None:
            # Use Inverse Kinematics to get the goal configuration of the robot
            self._goal_config = self.solve_pose_ik_problem(
                self.goal_pose,
                robot_joint_names=hardcoded_robot_joint_names,
            )

            return self._goal_config
        else:
            return self._goal_config

    @property
    def goal_pose(self) -> RigidTransform:
        """
        Description
        -----------
        Get the goal pose of the end effector frame.
        """
        # Setup

        # Algorithm
        if self._goal_pose is not None:
            return self._goal_pose
        elif (self._goal_pose is None) and (self._goal_config is None):
            # If we have no guidance on how to start, then we will use a default
            # goal_position = np.array([+0.0, 0.5, 0.7])
            # goal_orientation = RollPitchYaw(np.pi / 2.0, np.pi / 2.0, 0.0).ToQuaternion()

            # goal_position = np.array([-0.5, 0.0, 0.7])
            # goal_orientation = RollPitchYaw(0.0, np.pi *(0.0/ 2.0), 0.0).ToQuaternion()

            # self._goal_pose = RigidTransform(goal_orientation, goal_position)

            self._goal_pose = self.solve_forward_kinematics_problem_for_arm(
                np.array([
                    2.18266076, -6.26145905, -1.21521333, -3.56075315,  1.45027155,  5.12764245
                ])
            )

            return self._goal_pose
        elif (self._goal_pose is None) and (self._goal_config is not None):
            # If the goal configuration is given,
            # use the forward kinematics solver to get the goal pose
            self._goal_pose = self.solve_forward_kinematics_problem_for_arm(self._goal_config)
            return self._goal_pose
        else:
            raise ValueError(
                f"It appears that the starting configuration is set, but the starting pose is not. This is unexpected behavior!"
            )

    @property
    def id(self) -> ProductionID:
        return ProductionID.kChemLab3
        
    def initialize_pose_data(self):
        # Setup
        table_width, table_length, table_height = self.table_width, self.table_length, self.table_height

        # Set shelf pose
        if self.shelf_pose is None:
            self.pose_WorldShelf = RigidTransform(
                RollPitchYaw(0.0, 0.0, -np.pi/2.0).ToQuaternion(),
                np.array([0.0, 0.95, 0.66]),
            )

        # Define poses for the two halves of the table
        self.pose_WorldTableHalf1 = RigidTransform(
            RollPitchYaw(0.0, 0.0, 0.0).ToQuaternion(),
            np.array([0.0, 0.75, 0.0]),
        )
        self.pose_WorldTableHalf2 = RigidTransform(
            RollPitchYaw(0.0, 0.0, 0.0).ToQuaternion(),
            np.array([-table_width/2.0 + (table_width/3.0)*0.5, -0.2, 0.0]),
        )

        # Define pose of the starting point of the flask
        if self.pose_WorldFlask0 is None:
            table2_to_flask_translation0 = np.array([+0.0, 0.0, 0.1])
            table2_to_flask_orientation0 = RollPitchYaw(0.0, 0.0, 0.0).ToQuaternion()
            pose_Table2Flask0 = RigidTransform(
                table2_to_flask_orientation0,
                table2_to_flask_translation0,
            )
            self.pose_WorldFlask0 = self.pose_WorldTableHalf2.multiply(pose_Table2Flask0)

        # Define starting pose for the robot w.r.t. the flask
        flask_to_start_translation = np.array([+0.0, 0.0, 0.5])
        flask_to_start_orientation = RollPitchYaw(0.0, 0.0, 0.0).ToQuaternion()
        self.pose_FlaskStart = RigidTransform(
            flask_to_start_orientation,
            flask_to_start_translation,
        )

    def solve_forward_kinematics_problem_for_arm(
        self,
        robot_joint_positions: np.ndarray,
    ) -> RigidTransform:
        """
        Description
        -----------
        This method solves the forward kinematics problem for the UR10e arm
        by itself.
        """
        # Setup
        target_frame_name = "tool0"

        # Create shadow plant and populate it with the UR10e arm
        shadow_station = UR10eStation(
            time_step=self.time_step,
            meshcat_port_number=None,
            gripper_type=self.gripper_type,
        )
        shadow_station.Finalize()

        # Use station's plant to solve the forward kinematics problem
        shadow_plant = shadow_station.plant
        
        temp_context = shadow_plant.CreateDefaultContext()
        shadow_plant.SetPositions(
            temp_context,
            shadow_station.arm,
            robot_joint_positions,
        )

        # Get the pose of the end effector
        end_effector_pose = shadow_plant.EvalBodyPoseInWorld(
            temp_context,
            shadow_plant.GetBodyByName(target_frame_name),
        )

        return end_effector_pose

    @property
    def start_configuration(self):
        """
        Get the start configuration of the robot.
        In other words, the positions of the robots joints.
        This should be defined by the subclass.
        :return:
        """
        # Setup
        hardcoded_robot_joint_names = [
            'ur10e-test_shoulder_pan_joint_q',
            'ur10e-test_shoulder_lift_joint_q',
            'ur10e-test_elbow_joint_q',
            'ur10e-test_wrist_1_joint_q',
            'ur10e-test_wrist_2_joint_q',
            'ur10e-test_wrist_3_joint_q',
        ]

        # Algorithm
        if self._start_config is not None:
            return self._start_config
        elif self._start_pose is not None:
            # Use the start pose to get the start configuration
            # Using the IK solver (potentially buggy because default ik problem ignores obstacles)
            return self.solve_pose_ik_problem(
                self._start_pose,
                robot_joint_names=hardcoded_robot_joint_names,
            )
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )


    @property
    def start_pose(self) -> RigidTransform:
        """
        Get the start pose. This should be defined by the subclass.
        :return:
        """
        # Setup

        # Algorithm
        if not (self._start_pose is None):
            return self._start_pose
        elif (self._start_pose is None) and (self._start_config is None):
            # Define Start Pose
            pose_WorldStart = self.pose_WorldFlask0.multiply(self.pose_FlaskStart)
            self._start_pose = pose_WorldStart
            return self._start_pose
        elif (self._start_pose is None) and (self._start_config is not None):
            # Use the start configuration to get the starting pose
            # Using a "forward kinematics solver"
            self._start_pose = self.solve_forward_kinematics_problem_for_arm(self._start_config)
            return self._start_pose
        else:
            raise ValueError(
                f"Unexpected behavior. This should never happen."
            )



"""
Description:
    This file will contain utilities for quickly creating simple URDF files
    to be used in debugging and testing.
"""
# External Imports
from dataclasses import dataclass
from enum import IntEnum
import os
import xml.etree.ElementTree as ET

import numpy as np
from pathlib import Path
from pydrake.math import RigidTransform

# Internal Imports
from brom_drake.file_manipulation.urdf.shapes.shape_definition import ShapeEnum, ShapeDefinition
from brom_drake.file_manipulation.urdf.simple_writer.inertia_definition import InertiaDefinition
from brom_drake.directories import DEFAULT_BROM_MODELS_DIR

@dataclass
class SimpleShapeURDFDefinition:
    """
    *Description*

    A simple shape along with the optional and required options to fully
    specifiy it in a URDF.

    *Attributes*
    
    name: str
        The name of the shape.

    shape: ShapeDefinition
        The shape definition for the shape.

    color: np.ndarray, optional
        The RGBA color of the shape.

    create_collision: bool, optional
        Whether to create the collision elements of the urdf.

    mass: float, optional
        The mass of the shape.
        Defaults to 1.0.

    inertia: InertiaDefinition, optional
        The inertia of the shape.
        If not specified, a default inertia will be used.

    pose: RigidTransform, optional
        The pose of the shape.
        Defaults to identity.

    mu_static: float, optional
        The static friction coefficient of the shape.
        Defaults to 0.7.

    mu_dynamic: float, optional
        The dynamic friction coefficient of the shape.
        Defaults to 0.4.

    is_hydroelastic: bool, optional
        Whether to use hydroelastic collision for the shape.
        Defaults to True.
    """
    name: str
    shape: ShapeDefinition
    color: np.ndarray = None
    create_collision: bool = True # Whether to create the collision elements of the urdf
    mass: float = 1.0
    inertia: InertiaDefinition = None
    pose: RigidTransform = RigidTransform()
    mu_static: float = 0.7
    mu_dynamic: float = 0.4
    is_hydroelastic: bool = True # Whether to use hydroelastic collision for the shape

    def as_urdf(self) -> ET.Element:
        """
        *Description*

        Create the URDF for the simple shape using python's built-in xml library.

        *Returns*

        xml.etree.ElementTree.Element
            The root element of the desired URDF.
        """
        # Setup
        root = ET.Element(
            "robot",
            {
                "name": self.name + "_robot",
                "xmlns:drake": "http://drake.mit.edu",
            }
        )
        link = ET.SubElement(root, "link", {"name": self.base_link_name})

        # Add inertial elements to link
        self.add_inertial_elements_to(link)

        # Add visual elements to link
        self.add_visual_elements_to(link)

        # Add collision elements to link
        if self.create_collision:
            self.add_collision_elements_to(link)

        return root

    def add_inertial_elements_to(self, link_elt: ET.Element):
        """
        *Description*

        Add the inertial elements to the link element.
        
        *Parameters*

        link_elt: xml.etree.ElementTree.Element
            The ET.Element object for the link.

        *Returns*

        Nothing, but modifies the link_elt in place.
        """
        # Setup
        inertial_link = ET.SubElement(link_elt, "inertial")
        inertia_def = self.inertia
        if inertia_def is None:
            inertia_def = InertiaDefinition()

        # Create all child links for inertial elements
        self.add_origin_element_to(inertial_link)
        mass = ET.SubElement(inertial_link, "mass", {"value": f"{self.mass}"})
        inertia = ET.SubElement(
            inertial_link,
            "inertia",
            inertia_def.as_map(),
        )

        # Add inertial elements to link
        link_elt.append(inertial_link)

    def add_visual_elements_to(self, link_elt: ET.Element):
        """
        *Description*

        Add the visual elements to the link element.
        
        *Parameters*

        link_elt: xml.etree.ElementTree.Element
            The ET.Element object for the link.

        *Returns*

        Nothing, but modifies the link_elt in place.
        """
        # Setup
        visual_link = ET.SubElement(link_elt, "visual")
        color = self.color
        if color is None:
            color = np.array([0.0, 1.0, 0.0, 0.8])

        # Create all child links for visual elements
        self.add_origin_element_to(visual_link)
        geometry = ET.SubElement(visual_link, "geometry")

        # Add the shape to the geometry element
        shape = self.shape.add_geometry_to_element(geometry)

        # Add the material to the visual element
        material = ET.SubElement(visual_link, "material", {"name": "blue"})
        color = ET.SubElement(
            material,
            "color",
            {"rgba": f"{color[0]} {color[1]} {color[2]} {color[3]}"},
        )

    def add_collision_elements_to(self, link_elt: ET.Element):
        """
        *Description*

        Add the collision elements to the link element.

        *Parameters*

        link_elt: xml.etree.ElementTree.Element
            The ET.Element object for the link.

        *Returns*
        Nothing, but modifies the link_elt in place.
        """
        # Setup
        collision_link = ET.SubElement(link_elt, "collision")

        # Add pose to collision element
        self.add_origin_element_to(collision_link)
        geometry = ET.SubElement(collision_link, "geometry")

        # Add the shape to the geometry element
        self.shape.add_geometry_to_element(geometry)

        # Add the proximity properties to the collision element
        self.add_proximity_properties_to(collision_link)

    def add_proximity_properties_to(self, collision_elt: ET.Element):
        """
        *Description*

        Add the proximity properties to the collision element.

        *Parameters*

        collision_elt: xml.etree.ElementTree.Element
            The ET.Element object for the collision object.

        *Returns*

        Nothing, but modifies the collision_elt in place.
        """
        # Setup
        proximity_properties = ET.SubElement(collision_elt, "proximity_properties")
        
        # Create mu_static element
        mu_static= ET.SubElement(
            proximity_properties,
            "drake:mu_static",
            {"value": f"{self.mu_static}"}
            )

        # Create mu_dynamic element
        mu_dynamic = ET.SubElement(
            proximity_properties,
            "drake:mu_dynamic",
            {"value": f"{self.mu_dynamic}"}
            )
        
        # Create is_hydroelastic element
        hydroelastic = ET.SubElement(
            proximity_properties,
            "drake:rigid_hydroelastic",
        )

    def add_origin_element_to(self, target_element: ET.Element):
        """
        *Description*

        Add the origin element to the target element.
        
        *Parameters*

        target_element: xml.etree.ElementTree.Element
            The ET.Element object to which the origin element will be added.

        *Returns*
        Nothing, but modifies the target_element in place.
        """
        # Setup
        translation = self.pose.translation()
        rot_as_rpy = self.pose.rotation().ToRollPitchYaw()
        origin_elt = ET.SubElement(
            target_element,
            "origin",
            {
                "xyz": f"{translation[0]} {translation[1]} {translation[2]}",
                "rpy": f"{rot_as_rpy.roll_angle()} {rot_as_rpy.pitch_angle()} {rot_as_rpy.yaw_angle()}",
            },
        )

    @property
    def base_link_name(self) -> str:
        """
        *Description*

        The name of the base link.

        *Returns*

        str
            The name of the base link.
        """
        return self.name + "_base_link"

    def write_to_file(self, file_path: str = None) -> str:
        """
        *Description*

        Write the URDF to a file.

        *Parameters*

        file_path: str
            The path to the file where the URDF will be written.
        
        *Returns*
        
        str
            The path to the file where the URDF was written.
        """
        # Setup
        root = self.as_urdf()
        tree = ET.ElementTree(root)
        ET.indent(tree, space="\t", level=0)

        # Input Processing
        if file_path is None:
            file_path = DEFAULT_BROM_MODELS_DIR + "/shapes/" + self.name + ".urdf"

        # Create the directory if it doesn't exist
        fp_as_path = Path(file_path)
        fp_as_path.parent.mkdir(parents=True, exist_ok=True)

        # Write to file
        tree.write(file_path, xml_declaration=True)

        return file_path





from pydrake.common.value import AbstractValue
from pydrake.all import RigidTransform
from pydrake.systems.framework import Context, InputPort, LeafSystem

class PoseCompositionSystem(LeafSystem):
    """
    *Description*

    This LeafSystem receives two Pose inputs (i.e., AbstractValue objects
    containing RigidTransform values) and returns the composition of the two
    poses (i.e., an AbstractValue containing a RigidTransform value).

    *Diagram*

    The LeafSystem's input and output ports can be illustrated with the following block: ::

                            |---------------|
        pose_AB ----------->|               |
        (RigidTransform)    | Pose          |
                            | Composition   | ----> pose_AC
        pose_BC ----------->| System        |       (RigidTransform)
        (RigidTransform)    |               |
                            |---------------|
    """

    def __init__(self):
        LeafSystem.__init__(self)

        # Define Input Ports
        self._pose_AB_port = self.DeclareAbstractInputPort(
            "pose_AB",
            AbstractValue.Make(RigidTransform())
        )
        self._pose_BC_port = self.DeclareAbstractInputPort(
            "pose_BC",
            AbstractValue.Make(RigidTransform())
        )

        # Define Output Port
        self.DeclareAbstractOutputPort(
            "pose_AC",
            lambda: AbstractValue.Make(RigidTransform()),
            self.CalcOutputPose,
        )

    def CalcOutputPose(self, context: Context, output: AbstractValue):
        """
        *Description*

        This callback function computes the composed pose, ``pose_AC``,
        from the two inputs ``pose_AB`` and ``pose_BC``.
        """
        # Retrieve current input values
        pose_AB: RigidTransform = self.GetInputPort("pose_AB").Eval(context)
        pose_BC: RigidTransform = self.GetInputPort("pose_BC").Eval(context)
        
        assert type(pose_AB) is RigidTransform, \
            f"Expected pose_AB port to contain \"RigidTransform\" objects, but received type {type(pose_AB)}"
        
        assert type(pose_BC) is RigidTransform, \
            f"Expected pose_BC port to contain \"RigidTransform\" objects, but received type {type(pose_BC)}"

        # Compute Product and return it
        pose_AC = pose_AB.multiply(pose_BC)

        output.SetFrom(
            AbstractValue.Make(pose_AC)
        )

    def get_pose_AB_port(self) -> InputPort:
        return self._pose_AB_port

    def get_pose_BC_port(self) -> InputPort:
        return self._pose_BC_port


from brom_drake.productions.roles.role import Role, RolePortAssignment
from brom_drake.productions.roles.motion_planners.kinematic import kKinematicMotionPlanner

__all__ = [
    "Role",
    "kKinematicMotionPlanner",
    "RolePortAssignment",
]



from importlib import resources as impresources
import networkx as nx
import numpy as np
from pydrake.all import (
    AbstractValue,
    CollisionFilterDeclaration,
    ConstantValueSource,
    ConstantVectorSource,
    Context,
    Diagram,
    GeometrySet,
    MultibodyPlant,
    Parser,
    Quaternion,
)
from pydrake.math import RollPitchYaw, RigidTransform
from typing import Callable, List, Tuple

# Internal Imports
from brom_drake.directories import DEFAULT_BROM_MODELS_DIR
from brom_drake.control import (
    GripperTarget,
)
from brom_drake.file_manipulation.urdf import drakeify_my_urdf
from brom_drake.file_manipulation.urdf.shapes.box import BoxDefinition
from brom_drake.file_manipulation.urdf.simple_writer.urdf_definition import SimpleShapeURDFDefinition, \
    InertiaDefinition
from brom_drake.motion_planning.systems.open_loop_dispensers.open_loop_plan_dispenser import OpenLoopPlanDispenser
import brom_drake.robots as robots
from brom_drake.robots.gripper_type import GripperType
from brom_drake.stations.kinematic import UR10eStation as KinematicUR10eStation
from brom_drake.productions import ProductionID
from brom_drake.productions.roles import Role
from brom_drake.productions.types import KinematicMotionPlanningProduction
from brom_drake.utils import Performer, AddGround, MotionPlan


class ChemLab1(KinematicMotionPlanningProduction):
    """
    *Description*

    This production is the first in the chemistry lab series.
    It is used to test the motion planning capabilities of the robot
    in a chemistry lab setting with minimal constraints.
    
    *Parameters*

    time_step: float, optional
        The time step for the production, by default 1e-3.

    meshcat_port_number: int, optional
        The port number for meshcat visualization, by default 7001.

    plan_execution_speed: float, optional
        The speed at which to execute the motion plan, by default 0.2.

    pose_WorldBeaker: RigidTransform, optional
        The pose of the beaker in the world frame, by default None.

    table_length: float, optional
        The length of the table, by default 0.6.

    table_width: float, optional
        The width of the table, by default 2.0.

    table_height: float, optional
        The height of the table, by default 0.1.

    shelf_pose: RigidTransform, optional
        The pose of the shelf in the world frame, by default None.

    *Usage*

    To use this production, create an instance of the ChemLab1 class
    provide it with a motion planning function, and then build it. ::

        # Create the production
        production = ChemLab1()

        # Create a planner object which will be used to plan the motion
        config = RRTConnectPlannerConfig(
            steering_step_size=0.1,
            prob_sample_goal=0.30,
            max_iterations=int(1e5),
            convergence_threshold=1e-3,
            debug_flag=True,
        )
        planner2 = RRTConnectPlanner(
            production.arm,
            production.plant,
            production.scene_graph,
            config=config,
        )

        # To build the production, we only need to provide a planning function
        # (can come from anywhere, not just a BaseRRTPlanner object)
        diagram, diagram_context = production.easy_cast_and_build(
            planner2.plan,
            with_watcher=True,
        )

        print("Simulating...")

        # Simulate the diagram
        simulator = Simulator(diagram, diagram_context)
        simulator.set_target_realtime_rate(1.0)

        # Run simulation
        simulator.Initialize()
        simulator.AdvanceTo(0.1)
        planned_trajectory = production.plan_dispenser.planned_trajectory
        print(f"Expected end time of trajectory: {planned_trajectory.end_time()}")

        # Advance to the end of the simulation
        simulator.AdvanceTo(planned_trajectory.end_time()+1.0)
    """
    def __init__(
        self,
        time_step: float =1e-3,
        meshcat_port_number: int = 7000,
        plan_execution_speed: float = 0.2,
        table_length: float = 0.6,
        table_width: float = 2.0,
        table_height: float = 0.1,
        shelf_pose: RigidTransform = None,
        **kwargs,
    ):
        """
        *Description*
            
        Constructor for the ChemLab1Scene class.

        *Parameters*

        time_step: float, optional
            The time step for the production, by default 1e-3.

        meshcat_port_number: int, optional
            The port number for meshcat visualization, by default 7001.

        plan_execution_speed: float, optional
            The speed at which to execute the motion plan, by default 0.2.

        pose_WorldBeaker: RigidTransform, optional
            The pose of the beaker in the world frame, by default None.

        table_length: float, optional
            The length of the table, by default 0.6.

        table_width: float, optional
            The width of the table, by default 2.0.

        table_height: float, optional
            The height of the table, by default 0.1.

        shelf_pose: RigidTransform, optional
            The pose of the shelf in the world frame, by default None.
        
        """
        # Superclass constructor
        super().__init__(**kwargs)

        # Input Processing
        self.time_step = time_step
        self.meshcat_port_number = meshcat_port_number
        self.plan_execution_speed = plan_execution_speed
        self.table_height, self.table_width = table_height, table_width
        self.table_length = table_length

        # Set shelf pose
        self.shelf_pose = shelf_pose
        if self.shelf_pose is None:
            self.pose_WorldShelf = RigidTransform(
                RollPitchYaw(0.0, 0.0, -np.pi/2.0).ToQuaternion(),
                np.array([0.0, 0.75, 0.66]),
            )

        # Set Beaker Pose
        self.pose_WorldBeaker = RigidTransform(
            RollPitchYaw(np.pi/2.0, 0.0, 0.0).ToQuaternion(),
            np.array([-0.6, 0.45, 0.075]),
        )

        # Set Holder Pose
        self.pose_WorldHolder = RigidTransform(
            RollPitchYaw(np.pi/2.0, 0.0, 0.0).ToQuaternion(),
            np.array([self.table_width*0.5*0.7, 0.6+self.table_length/4., self.table_height-0.015]),
        )

        # Set station
        self.station = KinematicUR10eStation(
            time_step=self.time_step,
            meshcat_port_number=self.meshcat_port_number,
            gripper_type=GripperType.NoGripper,
        )
        self.arm = self.station.arm
        self.robot_model_idx_ = self.arm

        # self.gripper = self.station.gripper

        # Set Names of Plant and scene graph
        self.plant = self.station.plant
        self.scene_graph = self.station.scene_graph

        self.plant.set_name(f"ChemLab1_Production_Plant")
        self.scene_graph.set_name(f"ChemLab1_Production_SceneGraph")

        # Define placeholder variables for models
        self.test_tube_holder1 = None

    def add_supporting_cast(self):
        """
        *Description*
        
        This method adds all secondary cast members to the builder.
        The secondary cast members in the production are the:
        
        - Table, where the robot exists
        - Test Tube Holders
        - Component which share's the robot model reference
        - Motion Planning components (e.g., dispensers, etc.)
        - Start and Goal sources
        
        """
        # Call the superclass method
        super().add_supporting_cast()

        # Setup

        # Create the table and the items on it
        self.add_table()
        self.add_test_tube_holders()
        self.add_beaker()
        self.add_shelf()
        AddGround(self.plant)

        # Add the station
        self.builder.AddSystem(
            self.station,
        )

        self.station.Finalize()

        # Add the motion planning components
        self.add_robot_source_system()
        self.add_motion_planning_components()
        self.add_start_and_goal_sources_to_builder()

        # Add initial conditions for the arm
        self.initial_condition_manager.add_initial_configuration(
            model_instance_index=self.arm,
            configuration=self.start_configuration,
        )
        # The initial condition will be set for the plant when we build the production.

        print("Completed adding supporting cast members.")

    def add_beaker(self):
        """
        *Description*
        
        This method adds a model of a beaker to the production.
        The beaker is welded in place at the pose pose_WorldBeaker
        with respect to the world origin.

        The beaker's URDF is specified in the `brom_drake` robots
        package.
        """
        # Setup
        plant = self.plant
        urdf_file_path = str(
            impresources.files(robots) / "models/beaker/beaker.urdf"
        )

        # Use Drakeify my urdf to create the beaker
        new_beaker_urdf = drakeify_my_urdf(urdf_file_path)

        # Add the beaker to the plant
        model_idcs = Parser(plant=plant).AddModels(str(new_beaker_urdf))
        self.beaker_model_index = model_idcs[0]

        # Weld the beaker to the world frame
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("beaker_base_link", self.beaker_model_index),
            self.pose_WorldBeaker,
        )

    def add_motion_planning_components(self):
        """
        *Description*

        Add the motion planning components to the builder.
        """
        # Setup
        arm = self.arm
        n_actuated_dof = self.plant.num_actuated_dofs(arm) # Number of actuated DOF in arm

        # Add the Plan Dispenser and connect it to the station
        self.plan_dispenser = self.builder.AddSystem(
            OpenLoopPlanDispenser(n_actuated_dof, self.plan_execution_speed)
        )

        self.builder.Connect(
            self.plan_dispenser.GetOutputPort("point_in_plan"),
            self.station.GetInputPort("desired_joint_positions"),
        )

    def add_shelf(self):
        """
        *Description*

        Add a model of a shelf to the production's plant.
        The shelf's pydrake.multibody.tree.ModelInstanceIndex will be saved in the internal variable
        ``self.shelf_model_index``.
        The shelf is fixed at the pose ``self.pose_WorldShelf`` with respect to the world origin.
        """
        # Setup
        plant = self.plant
        urdf_file_path = str(
            impresources.files(robots) / "models/cupboard/cupboard.sdf"
        )

        # Add the shelf to the plant
        model_idcs = Parser(plant=plant).AddModels(urdf_file_path)
        self.shelf_model_index = model_idcs[0]

        # Weld the bookshelf to the world frame
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("cupboard_body", self.shelf_model_index),
            self.pose_WorldShelf,
        )

    def add_table(self):
        """
        *Description*
        
        This method adds a model of a table to the production's plant.
        The table will be a simple shape that the robot will interact with.
        The table is fixed in a pose with respect to the origin (pose chosen in this function).
        """
        # Setup

        # Create a new shape urdf
        table_defn = SimpleShapeURDFDefinition(
            name="table",
            shape=BoxDefinition(
                size=(self.table_width, self.table_length, self.table_height),
            ),
            mass=100.0, # kg
            inertia=InertiaDefinition(
                ixx=10.0,
                iyy=10.0,
                izz=10.0,
            ),
            color=np.array([0.1, 0.1, 0.1, 0.5]),
        )
        table_urdf_path = DEFAULT_BROM_MODELS_DIR + "/table/table.urdf"
        table_defn.write_to_file(table_urdf_path)

        # Add the table to the production
        table_model_index = Parser(self.plant).AddModels(table_urdf_path)[0]

        # Weld the table to the world frame
        table_pose = RigidTransform(
            RollPitchYaw(0.0, 0.0, 0.0).ToQuaternion(),
            np.array([0.0, 0.6, 0.0]),
        )
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName("table_base_link", table_model_index),
            table_pose,
        )

    def add_test_tube_holders(self):
        """
        *Description*
        
        This method adds the test tube holders to the production.
        The test tube holder is fixed at pose ``self.pose_WorldHolder``
        with respect to the world origin.
        """
        # Setup
        test_tube_urdf1 = str(
            impresources.files(robots) / "models/test_tube_holder/test_tube_holder.urdf"
        )

        # Use Drakeify my urdf to create the test tube holders
        new_test_tube_holder_urdf1 = drakeify_my_urdf(test_tube_urdf1)

        self.test_tube_holder1 = Parser(self.plant).AddModels(
            str(new_test_tube_holder_urdf1)
        )[0]

        # Weld the test tube holders to the world frame
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName("test_tube_holder_base_link", self.test_tube_holder1),
            self.pose_WorldHolder,
        )


    def add_cast_and_build(
        self,
        cast: List[Tuple[Role, Performer]] = [],
        with_watcher: bool = False,
    ) -> Tuple[Diagram, Context]:
        """
        *Description*
        
        Modifies the normal add_cast_and_build, so that
        we share the context of the plant with the appropriate
        parts of the system.

        In addition to using the parent class's implementation of add_cast_and_build, this:
        - Configures the collision filter to avoid spurious collisions between the cupboard's 
        various parts and itself
        - Insert the motion planner role into the diagram and connect it properly
        
        *Parameters*

        cast: List[Tuple[Role, Performer]], optional
            The main cast members that we would like to assign to the production.
            Default is None.
            TODO(Kwesi): Will using the default lead to errors? Should we even allow that?
        
        with_watcher: bool, optional
            A Boolean that determines whether to add a watcher to the diagram.
            Default is False.
        
        *Returns*

        diagram: pydrake.systems.framework.Diagram
            The diagram that represents the fully assembled cast, all connected together.

        diagram_context: pdrake.systems.framework.Context
            The context for the diagram (used by the diagram watcher to monitor the system, if defined).

        .. note::

            This method returns a `pydrake.systems.framework.Diagram` and its associated `pydrake.systems.framework.Context`.
            When constructing a Drake simulation of this, you should use these objects to create the `pydrake.systems.analysis.Simulator`.
            
        """
        diagram, diagram_context = super().add_cast_and_build(
            main_cast_members=cast,
            with_watcher=with_watcher,
        )

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
            self.station.arm_controller.plant, diagram_context,
        )

        for role_ii, performer_ii in cast:
            if role_ii.name == "OfflineMotionPlanner":
                performer_ii.set_internal_root_context(
                    diagram_context
                )

        return diagram, diagram_context
    
    def configure_collision_filter(self, scene_graph_context: Context):
        """
        *Description*
        
        This method configures the collision filter, so that:
        
        - self collisions between the shelf's pieces,
        
        are ignored during simulation.

        *Parameters*

        scene_graph_context: pydrake.systems.framework.Context
            The context of the scenegraph for the constructed production (i.e., the diagram
            for the built Production).
        """
        # Setup
        scene_graph = self.scene_graph
        shelf_model_index = self.shelf_model_index

        # Ignore self collisions of the shelf, by:
        # - Getting the model instance index of the shelf
        # - Collecting the Geometry IDs of all the geometries in the shelf
        shelf_geometry_ids = []
        for body_index in self.plant.GetBodyIndices(shelf_model_index):
            # Get the geometry IDs
            shelf_geometry_ids.extend(
                self.plant.GetCollisionGeometriesForBody(
                    self.plant.get_body(body_index)
                )
            )
            # print(self.plant.get_body(body_index))

        self.shelf_geometry_ids = shelf_geometry_ids

        # Apply the collision filter
        scene_graph.collision_filter_manager(scene_graph_context).Apply(
            CollisionFilterDeclaration().ExcludeWithin(
                GeometrySet(self.shelf_geometry_ids)
            )
        )

        # # Ignore collisions between the robot links
        # # TODO: Actually implement this for adjacent links? Or is this not necessary?
        # arm_geometry_ids = []
        # for body_index in self.plant.GetBodyIndices(self.arm):
        #     arm_geometry_ids.extend(
        #         self.plant.GetCollisionGeometriesForBody(
        #             self.plant.get_body(body_index)
        #         )
        #     )

        # self.arm_geometry_ids = arm_geometry_ids

        # # Apply the collision filter
        # scene_graph.collision_filter_manager(scene_graph_context).Apply(
        #     CollisionFilterDeclaration().ExcludeWithin(
        #         GeometrySet(self.arm_geometry_ids)
        #     )
        # )
    
    def easy_cast_and_build(
        self,
        planning_algorithm: Callable[
            [np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],
            Tuple[MotionPlan, bool],
        ],
        with_watcher: bool = False,
    ) -> Tuple[Diagram, Context]:
        """
        *Description*
        
        This function is used to easily cast and build the production.

        *Parameters*

        planning_algorithm: Callable[[np.ndarray, np.ndarray, Callable[[np.ndarray], bool]], Tuple[MotionPlan, bool]
            The algorithm that produces a path from the start configuration to goal configuration
            using the collision checker for the scene.

        with_watcher: bool, optional
            A Boolean that determines whether to add a watcher to the diagram.
            Default is False.
        
        *Returns*

        diagram: pydrake.systems.framework.Diagram
            The diagram that represents the fully assembled cast, all connected together.

        diagram_context: pdrake.systems.framework.Context
            The context for the diagram (used by the diagram watcher to monitor the system, if defined).

        .. note::

            This method returns a `pydrake.systems.framework.Diagram` and its associated `pydrake.systems.framework.Context`.
            When constructing a Drake simulation of this, you should use these objects to create the `pydrake.systems.analysis.Simulator`.
        """
        # Setup

        # Use Base class implementation to start
        diagram, diagram_context = super().easy_cast_and_build(
            planning_algorithm,
            with_watcher=with_watcher,
        )

        # print("completed cast and build.")

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
            self.station.arm_controller.plant, diagram_context,
        )

        self.performers[0].set_internal_root_context(
            diagram_context
        )

        return diagram, diagram_context

    @property
    def goal_configuration(self) -> np.ndarray:
        """
        *Description*

        Goal configuration of the robot in the production.

        Depending on the inputs, this may be computed the first time it is called.

        *Returns*

        q_goal: np.ndarray of shape (6,)
            The goal configuration of the robot.
        """
        # Setup
        hardcoded_robot_joint_names = [
            'ur10e-test_shoulder_pan_joint_q',
            'ur10e-test_shoulder_lift_joint_q',
            'ur10e-test_elbow_joint_q',
            'ur10e-test_wrist_1_joint_q',
            'ur10e-test_wrist_2_joint_q',
            'ur10e-test_wrist_3_joint_q',
        ]

        # Retrieve goal_configuraiton value
        if self._goal_config is None:
            beaker_to_goal_translation = np.array([+0.0, 0.2, 0.0]) 
            beaker_to_goal_orientation = RollPitchYaw(0., 0., 0.0).ToQuaternion()
            X_BeakerGoal = RigidTransform(  
                beaker_to_goal_orientation,
                beaker_to_goal_translation,
            )

            pose_WorldGoal = self.pose_WorldBeaker.multiply(X_BeakerGoal)

            # Use Inverse Kinematics to get the goal configuration of the robot
            self._goal_config = self.solve_pose_ik_problem(
                pose_WorldGoal,
                robot_joint_names=hardcoded_robot_joint_names,
            )

            return self._goal_config
        else:
            return self._goal_config

    @property
    def goal_pose(self) -> RigidTransform:
        """
        *Description*
        
        Get the goal pose of the end effector frame.

        *Returns*
        
        goal_pose: RigidTransform
            The goal pose of the robot.
        """
        # Setup

        # Algorithm
        if self._goal_pose is not None:
            # If the goal pose is already defined, return it
            return self._goal_pose
        elif (self._goal_pose is None) and (self._goal_config is None):
            # If we have no guidance on where goal is, then we will use the default goal pose
            beaker_to_goal_translation = np.array([+0.0, 0.2, 0.0]) 
            beaker_to_goal_orientation = RollPitchYaw(0., 0., 0.0).ToQuaternion()
            X_BeakerGoal = RigidTransform(  
                beaker_to_goal_orientation,
                beaker_to_goal_translation,
            )

            pose_WorldGoal = self.pose_WorldBeaker.multiply(X_BeakerGoal)
            self._goal_pose = pose_WorldGoal

            return self._goal_pose
        
        elif (self._goal_pose is None) and (self._goal_config is not None):
            # Use the goal configuration to get the goal pose
            # Using a "forward kinematics solver"
            self._goal_pose = self.solve_forward_kinematics_problem_for_arm(self._goal_config)
            return self._goal_pose
        
        else:
            raise ValueError(
                f"Unexpected behavior. This should never happen."
            )

    @property
    def id(self) -> ProductionID:
        """
        Always returns ``ProductionID.kChemLab1``.
        """
        return ProductionID.kChemLab1
        
    def solve_forward_kinematics_problem_for_arm(
        self,
        robot_joint_positions: np.ndarray,
        target_frame_name: str = "ft_frame",
    ) -> RigidTransform:
        """
        *Description*
        
        This method solves the forward kinematics problem for the UR10e arm
        by itself.

        *Parameters*

        robot_joint_positions: np.ndarray of shape (6,)
            The joint configuration of the robot in the production.

        target_frame_name: str, optional
            The frame that we compute the pose of with respect to the world origin frame.
            Default is "ft_frame"
        
        *Returns*
        
        pose_WorldTarget: RigidTransform
            The end effector pose of the robot expressed in the world frame.
        """
        # Setup

        # Create shadow plant and populate it with the UR10e arm
        shadow_station = KinematicUR10eStation(
            time_step=self.time_step,
            meshcat_port_number=None,
        )
        shadow_station.Finalize()

        # Use station's plant to solve the forward kinematics problem
        shadow_plant = shadow_station.plant
        
        temp_context = shadow_plant.CreateDefaultContext()
        shadow_plant.SetPositions(
            temp_context,
            shadow_station.arm,
            robot_joint_positions,
        )

        # Get the pose of the end effector
        end_effector_pose = shadow_plant.EvalBodyPoseInWorld(
            temp_context,
            shadow_plant.GetBodyByName(target_frame_name),
        )

        return end_effector_pose

    @property
    def start_configuration(self):
        """
        *Description*

        Get the start configuration.

        This may be computed using inverse kinematics on the first time it is requested,
        depending on the construction of the production.
        
        *Returns*

        q_start: np.ndarray of shape(6,)
            The starting configuration of the production's robot.
        """
        # Setup
        hardcoded_robot_joint_names = [
            'ur10e-test_shoulder_pan_joint_q',
            'ur10e-test_shoulder_lift_joint_q',
            'ur10e-test_elbow_joint_q',
            'ur10e-test_wrist_1_joint_q',
            'ur10e-test_wrist_2_joint_q',
            'ur10e-test_wrist_3_joint_q',
        ]

        # Algorithm
        if self._start_config is not None:
            return self._start_config
        elif self._start_pose is not None:
            # Use the start pose to get the start configuration
            # Using the IK solver (potentially buggy because default ik problem ignores obstacles)
            return self.solve_pose_ik_problem(
                self._start_pose,
                robot_joint_names=hardcoded_robot_joint_names,
            )
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )

    @property
    def start_pose(self) -> RigidTransform:
        """
        *Description*
        
        Get the start pose.
        
        *Returns*

        pose_WorldStart: RigidTransform
            The start pose of the robot's end effector frame with respect to the world frame.
        """
        # Setup

        # Algorithm
        if self._start_pose is not None:
            # If the start pose is already defined, return it
            return self._start_pose
        
        elif (self._start_pose is None) and (self._start_config is None):
            # If we have no guidance on how to start, then we will use the default start pose
            holder_to_start_translation = np.array([+0.0, 0.2, 0.025])
            holder_to_start_orientation = Quaternion(1, 0, 0, 0)
            X_HolderStart = RigidTransform(
                holder_to_start_orientation,
                holder_to_start_translation,
            )
            pose_WorldStart = self.pose_WorldHolder.multiply(X_HolderStart)
            self._start_pose = pose_WorldStart
            return self._start_pose
        
        elif (self._start_pose is None) and (self._start_config is not None):
            # Use the start configuration to get the starting pose
            # Using a "forward kinematics solver"
            self._start_pose = self.solve_forward_kinematics_problem_for_arm(self._start_config)
            return self._start_pose
        
        else:
            raise ValueError(
                f"Unexpected behavior. This should never happen."
            )




from dataclasses import dataclass
from enum import IntEnum
from typing import List

class MeshReplacementStrategy(IntEnum):
    """
    *Description*

    The different strategies for replacing mesh files.

    Options are:
        kDoNotReplace
        kWithObj
        kWithMinimalEnclosingCylinder
        kWithConvexDecomposition

    *Usage*

    .. code:: python

        from brom_drake.file_manipulation.urdf.drake_ready_urdf_converter.config import MeshReplacementStrategy
        # or
        # from brom_drake.all import MeshReplacementStrategy

        # Selecting the strategy to replace meshes with OBJ files
        strategy = MeshReplacementStrategy.kWithObj

    """
    kDoNotReplace = 0 
    kWithObj = 1
    kWithMinimalEnclosingCylinder = 2
    kWithConvexDecomposition = 3

@dataclass(frozen=True)
class MeshReplacementStrategies:
    """
    *Description*
    
    A dataclass that specifies the different strategies for replacing mesh files in:
    - the <collision> section of the URDF
    - the <visual> section of the URDF
    """
    collision_meshes: MeshReplacementStrategy = MeshReplacementStrategy.kWithObj
    visual_meshes: MeshReplacementStrategy = MeshReplacementStrategy.kWithObj

@dataclass
class DrakeReadyURDFConverterConfig:
    """
    *Description*
    
    A dataclass that specifies how the DrakeReadyURDFConverter
    should convert the URDF file.

    *Attributes*
    
    output_urdf_file_path: str
        The path where the converted URDF file will be saved. If None, it defaults to the models directory.
        
    coacd_log_level: str
        The log level for the coacd tool. Options are "off", "info", "error".
    """
    output_urdf_file_path: str = None
    overwrite_old_logs: bool = False
    overwrite_old_models: bool = False
    log_file_name: str = "conversion.log"
    mesh_replacement_strategies: MeshReplacementStrategies = MeshReplacementStrategies()
    add_missing_actuators: bool = True
    replace_colors_with: List[float] = None
    coacd_log_level: str = "info"


from .config import MeshReplacementStrategy, DrakeReadyURDFConverterConfig
from .converter import DrakeReadyURDFConverter
from .util import (
    find_mesh_file_path_in,
    tree_contains_transmission_for_joint,
    create_transmission_element_for_joint,
)

__all__ = [
    'create_transmission_element_for_joint',
    'DrakeReadyURDFConverter',
    'DrakeReadyURDFConverterConfig',
    'find_mesh_file_path_in',
    'MeshReplacementStrategy',
    'tree_contains_transmission_for_joint',
]


from brom_drake.utils.type_checking import is_rigid_transform
from pydrake.systems.framework import OutputPort, PortDataType
from typing import Any

def assert_port_is_supported(output_port: OutputPort):
    """
    *Description*
    
    Checks to see if the port is of the correct type for plotting.

    """
    # Algorithm
    if output_port.get_data_type() == PortDataType.kVectorValued:
        return
    
    # If the output port is kAbstractValued, then there are a specific
    # instantiations that we support.

    # Collect the example value for the port and then check that value
    example_allocation = output_port.Allocate()
    example_value = example_allocation.get_value()

    assert_abstract_value_ports_type_is_supported(example_value)

def assert_abstract_value_ports_type_is_supported(output_port_value: Any):
    # Check to see if AbstractValue port contains RigidTransform
    if is_rigid_transform(output_port_value):
        return
    elif type(output_port_value) == bool: # if the output_value is a boolean
        return

    # Check to see if the object is a list of one of the supported types
    if type(output_port_value) is list:
        # Then check every element of the list
        for example_elt in output_port_value:
            assert_abstract_value_ports_type_is_supported(
                output_port_value=example_elt
            )

        # If all of the above checks passed,
        # then return! We are good!
        return        

    # Raise error otherwise
    raise create_port_value_type_error(output_port_value)


def create_port_value_type_error(example_value: Any) -> ValueError:
    """
    *Description*
    
    Creates an error message for the port value type.
    
    *Parameters*
    
    example_value: Any
        The type of value in the problematic output port.

    *Returns*
    
    error_out: ValueError
        The error message.
    """
    # Return
    return ValueError(
        f"This watcher only supports output ports that are:\n" +
        f"- Vector valued ports (i.e., of type {PortDataType.kVectorValued}.\n" +
        f"- Abstract valued ports containing:\n" +
        f"  + RigidTransforms\n" +
        f"  + Booleans\n" +
        f"  + list[T], where T is a type from the above list."
        f"Received port of type {PortDataType.kAbstractValued} with underlying type {type(example_value)}."
    )


from brom_drake.PortWatcher.port_watcher_options import (
    PortFigureArrangement,
    FigureNamingConvention,
    PortWatcherPlottingOptions, 
)
from dataclasses import dataclass
from pathlib import Path
from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import OutputPort
from pydrake.systems.primitives import VectorLogSink
from typing import List

@dataclass
class PortWatcherFileManager:
    """
    *Description*
    
    This class manages file paths and directories for saving data
    collected by the PortWatcher system.
    """
    base_directory: Path
    plotting_options: PortWatcherPlottingOptions
    raw_data_file_format: str = "npy"

    @staticmethod
    def compute_safe_system_name(system_name: str) -> str:
        """
        *Description*
        
        This function returns a filesystem-safe version of the system name.

        *Returns*

        safe_system_name: str
            The filesystem-safe version of the system name.
        """
        # First, let's check to see how many "/" exist in the name
        slash_occurences = [i for i, letter in enumerate(system_name) if letter == "/"]
        if len(slash_occurences) > 0:
            system_name = system_name[slash_occurences[-1] + 1:]  # truncrate string based on the last slash
        
        # Second, replace all spaces with underscores
        system_name = system_name.replace(" ", "_")

        return system_name

    def compute_path_for_each_figure(
        self,
        output_port: OutputPort,
        associated_log_sink: VectorLogSink,
        port_component_name: str = None
    ) -> List[Path]:
        """
        *Description*
        
        Computes the names of all of the figures that will be produced for


        *Returns*
        
        figure_paths_out: List[Path]
            The paths of all of the figures that will be produced by
            this PortWatcherPlotter object.
        """
        # Setup
        plotting_options = self.plotting_options

        # Create the figure paths based on the naming convention given to the
        # PortWatcherPlotter.
        match plotting_options.figure_naming_convention:
            case FigureNamingConvention.kFlat:
                return self.figure_names_under_flat_convention(output_port, associated_log_sink, port_component_name)
            
            case FigureNamingConvention.kHierarchical:
                return self.figure_names_under_hierarchical_convention(output_port, associated_log_sink, port_component_name)
            
            case _:
                raise NotImplementedError(
                    f"Invalid figure naming convention for figure_names(): {plotting_options.figure_naming_convention}."
                )


    def figure_names_under_flat_convention(
        self,
        output_port: OutputPort,
        associated_log_sink: VectorLogSink,
        port_component_name: str = None
    ) -> List[Path]:
        """
        *Description*
        
        Returns the names associated with each figure that this port will
        generate assuming we are under the kFlat convention.

        *Arguments*

        output_port: OutputPort
            The output port for which we are generating figure names.
            We can extract the system name, port name, and port size from this.
            
        port_component_name: str
            A "sub-component" of the port that we wish to give a unique name in the
            figures.
        
        *Returns*
        
        figure_names: List[Path]
            List of paths where each path is a file name for an associated figure.
        """
        # Setup
        plotting_options = self.plotting_options
        format = plotting_options.file_format
        plot_dir = self.plot_dir

        # The naming also depends on the arrangement of the plots
        # (i.e., if there is one plot per port, or one plot per dimension)
        system = output_port.get_system()
        system_name = system.get_name()
        safe_system_name = self.compute_safe_system_name(system_name)

        port_name = output_port.get_name()
        log_sink_size = associated_log_sink.get_input_port().size()

        match plotting_options.plot_arrangement:
            case PortFigureArrangement.OnePlotPerPort:
                if port_component_name is None:
                    # If there is no sub-component name, then we just
                    # create the file in the main plot directory
                    return [
                        plot_dir / f"system_{safe_system_name}_port_{port_name}.{format}"
                    ]
                else:
                    # If there is a sub-component name, then we will
                    # create a sub-directory for it
                    return [
                        plot_dir / f"system_{safe_system_name}_port_{port_name}" / f"{port_component_name}.{format}"
                    ]

            case PortFigureArrangement.OnePlotPerDim:

                if port_component_name is None:
                    # If there is no sub-component name, then we just
                    # create the files in the main plot directory
                    return [
                        plot_dir / f"system_{safe_system_name}_port_{port_name}_dim{ii}.{format}"
                        for ii in range(log_sink_size)
                    ]
                else:
                    # If there is a sub-component name, then we will
                    # create a sub-directory for it
                    return [
                        plot_dir / f"system_{safe_system_name}_port_{port_name}" / f"{port_component_name}_dim{ii}.{format}"
                        for ii in range(log_sink_size)
                    ]
            case _:
                raise NotImplementedError(
                    f"Invalid plot arrangement for figure naming convention {plotting_options.figure_naming_convention}: {plotting_options.plot_arrangement}."
                )

    def figure_names_under_hierarchical_convention(
        self,
        output_port: OutputPort,
        associated_log_sink: VectorLogSink,
        port_component_name: str = None,
    ) -> List[Path]:
        """
        *Description*
        
        Returns the names associated with each figure that this port will
        generate assuming we are under the kHierarchical convention.
        
        *Parameters*
        
        output_port: OutputPort
            The output port for which we are generating figure names.
            We can extract the system name, port name, and port size from this.

        port_component_name: str
            A "sub-component" of the port that we wish to give a unique name in the figures.

        *Returns*
        
        paths_out: List[Path]
            Each path in this list is a file path for an associated figure.
        """
        # Setup
        plotting_options = self.plotting_options
        format = plotting_options.file_format

        # Compute the figure paths based on the arrangement of the plots
        # (i.e., if there is one plot per port, or one plot per dimension
        system = output_port.get_system()
        system_name = system.get_name()
        safe_system_name = self.compute_safe_system_name(system_name)

        port_name = output_port.get_name()
        
        log_sink_size = associated_log_sink.get_input_port().size()

        if plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            if port_component_name is None:
                # If there is no sub-component name, then we just
                # create the file in the main plot directory
                return [
                    self.plot_dir / f"system_{safe_system_name}/port_{port_name}.{format}"
                ]
            else:
                # If there is a sub-component name, then we will
                # create a sub-directory for it
                return [
                    self.plot_dir / f"system_{safe_system_name}/port_{port_name}/{port_component_name}.{format}"
                ]

        elif plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerDim:
            if port_component_name is None:
                # If there is no sub-component name, then we just
                # create the files in the main plot directory
                return [
                    self.plot_dir / f"system_{safe_system_name}/port_{port_name}/dim_{self.name_of_data_at_index(ii, output_port, associated_log_sink, remove_spaces=True)}.{format}"
                    for ii in range(log_sink_size)
                ]
            else:
                # If there is a sub-component name, then we will
                # create a sub-directory for it
                return [
                    self.plot_dir / f"system_{safe_system_name}/port_{port_name}/{port_component_name}/dim_{self.name_of_data_at_index(ii, output_port, associated_log_sink, remove_spaces=True)}.{format}"
                    for ii in range(log_sink_size)
                ]

        else:
            raise NotImplementedError(
                f"Invalid plot arrangement for figure naming convention {plotting_options.figure_naming_convention}: {plotting_options.plot_arrangement}."
            )

    def name_of_data_at_index(
        self,
        dim_index: int,
        target_port: OutputPort,
        associated_log_sink: VectorLogSink,
        remove_spaces: bool = False,
    ) -> str:
        """
        *Description*
        
        Returns the name of the data which is in index dim_index
        of this vector-valued port.

        TODO(kwesi): Consider moving this to its own utility file, outside of
        the file manager.
        
        *Parameters*
        
        self : PortWatcherPlotter
            The PortWatcherPlotter object.
        dim_index : int
            The index of the data in the port.
        remove_spaces : bool
            Whether to remove spaces from the name.
        
        *Returns*
        
        name_of_data: str
            The name of the data at index dim_index.
        """
        # Setup
        plotting_options = self.plotting_options

        n_dims_sink = associated_log_sink.get_input_port().size()
        system = target_port.get_system()

        # Input Processing
        if dim_index >= n_dims_sink:
            raise ValueError(
                f"dim_index ({dim_index}) is greater than the number of dimensions in the port ({n_dims_sink})."
            )

        # Default name
        name = f"Dim #{dim_index}"

        # If we are using the OnePlotPerPort config, then use the default name
        if plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            return name

        # Otherwise, try to get a better name
        # - For MultibodyPlants, we can get specific names for certain dimensions of the "state" output port

        if type(system) is MultibodyPlant:
            # The multi-body plant has names for specific ports
            if target_port.get_name() == "state":
                # We can get the names of the state
                state_names = system.GetStateNames()
                name = state_names[dim_index]
        
        # Filter our spaces, if requested
        if remove_spaces:
            name = name.replace(" ", "_")

        # Return name!
        return name

    @property
    def plot_dir(self) -> Path:
        """
        *Description*
        
        This function returns the directory where the plots will be saved.

        *Returns*
        
        plot_dir: Path
            The directory where the plots will be saved.
        """
        return self.base_directory / "plots"

    @property
    def raw_data_dir(self) -> Path:
        """
        *Description*
        
        This function returns the directory where the raw data will be saved.

        *Returns*
        
        raw_data_dir: Path
            The directory where the raw data will be saved.
        """
        return self.base_directory / "raw_data"

    def raw_data_file_path(
        self,
        system_name: str,
        port_name: str,
        port_component_name: str = None
    ) -> Path:
        """
        *Description*
        
        This function returns the file name for saving raw data.

        *Parameters*

        port_component_name: str
            A "sub-component" of the port that we wish to give a unique name in the data.

        *Returns*

        raw_data_file_name: Path
            The file name for saving raw data.
        """
        safe_system_name = self.compute_safe_system_name(system_name)

        if port_component_name is None:    
            return self.raw_data_dir / f"system_{safe_system_name}_port_{port_name}_data.{self.raw_data_file_format}"
        else:
            return self.raw_data_dir / f"system_{safe_system_name}_port_{port_name}" / f"{port_component_name}.{self.raw_data_file_format}"

    def time_data_file_path(self, system_name: str, port_name: str) -> Path:
        """
        *Description*
        
        This function returns the file name for saving time data.

        *Returns*

        time_data_file_path: Path
            The file name for saving time data.
        """
        safe_system_name = self.compute_safe_system_name(system_name)
        return self.raw_data_dir / f"system_{safe_system_name}_port_{port_name}_times.{self.raw_data_file_format}"
    


"""
gripper_controller.py
Description:

        This file defines the GripperController class.
"""
# External Imports
import importlib.resources as impresources

from pydrake.common.value import AbstractValue
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import JacobianWrtVariable
from pydrake.systems.framework import BasicVector, LeafSystem, Context

import os

import numpy as np

# Internal Imports
from .gripper_target import GripperTarget
from brom_drake.robots.gripper_type import GripperType

from brom_drake import robots


# Constants
package_dir = os.path.dirname(os.path.abspath(__file__))


class GripperController(LeafSystem):
    """
    *Description*

    A simple gripper controller with two modes: position and velocity.
    Both modes are essentially simple PD controllers.

    *Diagram*

    The GripperController LeafSystem can be represented with the following block diagram. ::

                                -------------------------
                                |                       |
                                |                       |
        gripper_target -------> |   GripperController   | ---> applied_gripper_torques
        gripper_target_type --> |                       |
                                |                       | ---> measured_gripper_position
                                |                       | ---> measured_gripper_velocity
        gripper_state --------> |                       |
                                |                       |
                                |                       |
                                -------------------------

    *Parameters*

    gripper_type: GripperType
        The type of gripper being controlled.

    Kp: np.ndarray, optional
        The proportional gain matrix for the PD controller.
        By default, this is set based on the gripper type.

    Kd: np.ndarray, optional
        The derivative gain matrix for the PD controller.
        By default, this is set based on the gripper type.
    """
    def __init__(self, gripper_type: GripperType, Kp: np.ndarray = None, Kd: np.ndarray = None):
        """
        *Description*
        
        Constructor for the GripperController class.
        :param gripper_type: The type can be "Robotiq_2f_85" or "NoGripper"
        """
        LeafSystem.__init__(self)
        self.type = gripper_type
        self.context = None

        # State input port size depends on what type of gripper we're using
        if self.type == GripperType.Robotiq_2f_85:
            state_size = 12

            # We'll create a simple model of the gripper which is welded to the floor.
            # This will allow us to compute the distance between fingers.
            self.plant = MultibodyPlant(time_step=1.0) # time step doesn't matter
            gripper_urdf = str(
                impresources.files(robots) / "models/robotiq/2f_85_gripper/urdf/robotiq_2f_85.urdf"
            )
            self.gripper = Parser(plant=self.plant).AddModels(gripper_urdf)[0]
            self.plant.WeldFrames(
                self.plant.world_frame(),
                self.plant.GetFrameByName("robotiq_arg2f_base_link"),
            )
            # Finalize the plant used for the gripper
            self.plant.Finalize()
            self.context = self.plant.CreateDefaultContext()

        else:
            raise RuntimeError("Invalid gripper type: %s" % self.type)

        # Declare input ports
        self.target_port = self.DeclareVectorInputPort(
            "gripper_target",
            BasicVector(1)
        )
        self.target_type_port = self.DeclareAbstractInputPort(
            "gripper_target_type",
            AbstractValue.Make(GripperTarget.kPosition),
        )
        self.state_port = self.DeclareVectorInputPort(
            "gripper_state",
            BasicVector(state_size),
        )

        # Declare output ports
        self.DeclareVectorOutputPort(
            "applied_gripper_torque",
            BasicVector(self.plant.num_actuators()),
            self.CalcGripperTorque,
        )
        self.DeclareVectorOutputPort(
            "measured_gripper_position",
            BasicVector(1),
            self.CalcGripperPosition,
            {self.time_ticket()}    # indicate that this doesn't depend on any inputs,
        )                           # but should still be updated each timestep
        self.DeclareVectorOutputPort(
            "measured_gripper_velocity", BasicVector(1), self.CalcGripperVelocity,
            {self.time_ticket()}    # indicate that this doesn't depend on any inputs,
        )                           # but should still be updated each timestep

        # Create the gain values
        self.Kp, self.Kd = None, None
        self.initialize_gains(Kp, Kd)

    def CalcGripperPosition(self, context: Context, output: BasicVector):
        """
        *Description*

        Calculate the gripper position from the state input port.

        *Parameters*

        context: pydrake.systems.framework.Context
            The current context of the system.

        output: pydrake.systems.framework.BasicVector
            The output vector to store the gripper position.
        """
        state = self.state_port.Eval(context)

        if self.type == "hande":
            width = 0.03
        elif self.type == GripperType.Robotiq_2f_85:  #2f_85
            width = 0.06
        else:
            raise NotImplementedError("Gripper type %s does not have CalcGripperPosition() implemented" % self.type)

        # Send a single number to match the hardware
        both_finger_positions = self.ComputePosition(state)
        net_position = 1/width* np.mean(both_finger_positions)

        output.SetFromVector([net_position])

    def CalcGripperVelocity(self, context, output):
        """
        *Description*
        
        Calculate the gripper velocity from the state input port.
        
        *Parameters*
        
        context: pydrake.systems.framework.Context
            The current context of the system.
            
        output: pydrake.systems.framework.BasicVector
            The output vector to store the gripper velocity.
        """
        state = self.state_port.Eval(context)

        if self.type == "hande":
            width = 0.03
        elif self.type == GripperType.Robotiq_2f_85:  # 2f_85
            width = 0.06
        else:
            raise NotImplementedError("Gripper type %s does not have CalcGripperVelocity() implemented" % self.type)

        # Send a single number to match the hardware
        both_finger_velocity = self.ComputeVelocity(state)
        net_velocity = 1 / width * np.mean(both_finger_velocity)

        output.SetFromVector([net_velocity])

    def ComputePosition(self, state: np.ndarray) -> np.ndarray:
        """
        *Description*

        Compute the gripper position from state data.
        This is especially useful for the 2F-85 gripper, since the
        state does not map neatly to the finger positions.

        *Parameters*

        state: np.ndarray
            The state vector of the gripper.

        *Returns*

        finger_position: np.ndarray of shape (2,)
            The positions of the gripper fingers.
        """
        # Setup
        finger_position = None

        # Get the finger positions
        if self.type == GripperType.Robotiq_2f_85:
            # For the more complex 2F-85 gripper, we need to do some kinematics
            # calculations to figure out the gripper position
            self.plant.SetPositionsAndVelocities(self.context, state)

            right_finger = self.plant.GetFrameByName("right_inner_finger_pad")
            left_finger = self.plant.GetFrameByName("left_inner_finger_pad")
            base = self.plant.GetFrameByName("robotiq_arg2f_base_link")

            X_lf = self.plant.CalcRelativeTransform(self.context,
                                                    left_finger,
                                                    base)
            X_rf = self.plant.CalcRelativeTransform(self.context,
                                                    right_finger,
                                                    base)

            lf_pos = -X_lf.translation()[1]
            rf_pos = -X_rf.translation()[1]

            finger_position = np.array([lf_pos, rf_pos])

        else:
            raise NotImplementedError("Gripper type %s does not have ComputePosition() implemented" % self.type)

        return finger_position

    def ComputeVelocity(self, state: np.ndarray) -> np.ndarray:
        """
        **Description**

        Compute the gripper velocity from state data.
        This is especially useful for the 2F-85 gripper, since the
        state does not map neatly to the finger positions.

        **Parameters**

        state: np.ndarray
            The state vector of the gripper.

        **Returns**

        finger_velocity: np.ndarray of shape (2,)
            The velocities of the gripper fingers.
        """
        if self.type == GripperType.Robotiq_2f_85:
            # For the more complex 2F-85 gripper, we need to do some kinematics
            self.plant.SetPositionsAndVelocities(self.context, state)
            v = state[-self.plant.num_velocities():]

            right_finger = self.plant.GetFrameByName("right_inner_finger_pad")
            left_finger = self.plant.GetFrameByName("left_inner_finger_pad")
            base = self.plant.GetFrameByName("robotiq_arg2f_base_link")

            J_lf = self.plant.CalcJacobianTranslationalVelocity(
                self.context, JacobianWrtVariable.kV, left_finger, np.zeros(3), base, base,
            )
            J_rf = self.plant.CalcJacobianTranslationalVelocity(
                self.context, JacobianWrtVariable.kV, right_finger, np.zeros(3), base, base,
            )

            lf_vel = -(J_lf @ v)[1]
            rf_vel = (J_rf @ v)[1]

            finger_velocity = np.array([lf_vel, rf_vel])

        else:
            raise NotImplementedError("Gripper type %s does not have ComputeVelocity() implemented" % self.type)

        return finger_velocity

    def CalcGripperPosition(self, context: Context, output: BasicVector):
        """
        *Description*

        Callback function for the `measured_gripper_position` output port.
        This method computes the gripper position based on the current state input.

        *Parameters*

        context: pydrake.systems.framework.Context
            The current context of the system.

        output: pydrake.systems.framework.BasicVector
            The output vector to store the gripper position.
        """
        state = self.state_port.Eval(context)

        if self.type == "hande":
            width = 0.03
        elif self.type == GripperType.Robotiq_2f_85:  #2f_85
            width = 0.06
        else:
            raise NotImplementedError("Gripper type %s does not have CalcGripperPosition() implemented" % self.type)

        # Send a single number to match the hardware
        both_finger_positions = self.ComputePosition(state)
        net_position = 1/width* np.mean(both_finger_positions)

        output.SetFromVector([net_position])

    def CalcGripperVelocity(self, context, output):
        state = self.state_port.Eval(context)

        if self.type == "hande":
            width = 0.03
        elif self.type == GripperType.Robotiq_2f_85:  # 2f_85
            width = 0.06
        else:
            raise NotImplementedError("Gripper type %s does not have CalcGripperVelocity() implemented" % self.type)

        # Send a single number to match the hardware
        both_finger_velocity = self.ComputeVelocity(state)
        net_velocity = 1 / width * np.mean(both_finger_velocity)

        output.SetFromVector([net_velocity])

    def CalcGripperTorque(self, context, output):
        # Setup
        state = self.state_port.Eval(context)
        target = self.target_port.Eval(context)
        target_type = self.target_type_port.Eval(context)

        width = self.gripper_width

        # Get PD Gains
        Kp, Kd = self.Kp, self.Kd

        # Collect State of gripper
        finger_position = self.ComputePosition(state)
        finger_velocity = self.ComputeVelocity(state)

        # Set target positions and velocities based on the current control mode
        if target_type == GripperTarget.kPosition:
            target = width - width * target * np.ones(2)
            target_finger_position = target
            target_finger_velocity = np.zeros(2)
        elif target_type == GripperTarget.kVelocity:
            target_finger_position = finger_position
            target_finger_velocity = -width * target
        else:
            raise RuntimeError("Invalid gripper target type: %s" % target_type)

        # Determine applied torques with PD controller
        position_err = target_finger_position - finger_position
        velocity_err = target_finger_velocity - finger_velocity
        tau = -Kp @ (position_err) - Kd @ (velocity_err)

        output.SetFromVector(tau)

    def initialize_gains(self, Kp: np.ndarray, Kd: np.ndarray):
        """
        *Description*
        
        Initialize the gains for the gripper controller.

        *Parameters*
        
        Kp: np.ndarray
            The proportional gain

        Kd: np.ndarray
            The derivative gain
        """
        # Handle proportional gain
        if Kp is not None:
            self.Kp = Kp
        else:
            # Gain depends on gripper type
            if self.type == GripperType.Robotiq_2f_85:
                self.Kp = 10 * np.eye(2)
            elif self.type == "hande":
                self.Kp = 100 * np.eye(2)
            else:
                raise NotImplementedError("Gripper type %s does not have Kp initialized" % self.type)

        # Handle derivative gain
        if Kd is not None:
            self.Kd = Kd
        else:
            # Gain depends on gripper type
            if self.type == GripperType.Robotiq_2f_85:
                self.Kd = 2 * np.sqrt(0.01 * self.Kp)
            elif self.type == "hande":
                self.Kd = 2 * np.sqrt(self.Kp)
            else:
                raise NotImplementedError("Gripper type %s does not have Kd initialized" % self.type)
    
    @property
    def gripper_width(self) -> float:
        """
        *Description*
        
        Return the gripper width.
        
        *Returns*

        width: float
            The width of the gripper.
        """
        if self.type == GripperType.Robotiq_2f_85:
            return 0.06
        elif self.type == "hande":
            return 0.03
        else:
            raise NotImplementedError("Gripper type %s does not have gripper_width implemented" % self.type)


from dataclasses import dataclass
import xml.etree.ElementTree as ET

# Internal Import
from .shape_definition import ShapeDefinition, ShapeEnum

@dataclass
class SphereDefinition(ShapeDefinition):
    """
    *Description*

    Defines a sphere shape as per URDF specifications.

    *Attributes*

    radius: float
        The radius of the sphere.
    """
    radius: float

    def add_geometry_to_element(self, target_element: ET.Element) -> ET.Element:
        """
        *Description*
        
        Adds the sphere geometry as a sub-element to the given target element.
        
        This method is an implementation of the abstract method defined in the
        ShapeDefinition base class.

        *Parameters*

        target_element: xml.etree.ElementTree.Element
            The target XML element to which the sphere geometry will be added.

        *Returns*

        xml.etree.ElementTree.Element
            The element `target_element` with the sphere geometry added as a sub-element.
        """
        return ET.SubElement(
            target_element,
            "sphere",
            {"radius": f"{self.radius}"},
        )

    @property
    def type(self) -> ShapeEnum:
        """
        *Description*
        
        Returns the ShapeEnum type corresponding to a sphere.

        This method is an implementation of the abstract method defined in the
        ShapeDefinition base class.

        *Returns*

        ShapeEnum
            The ShapeEnum type for a sphere.
        """
        return ShapeEnum.kSphere


from dataclasses import dataclass

@dataclass
class Configuration:
    """
    *Description*

    A dataclass that specifies the configuration for the RigidTransformToVectorSystem.

    *Attributes*

    name: str
        The name of the system.

    output_format: str, optional
        The format of the output vector. Options are:

        - "vector_xyz_quat(wxyz)"
        - "vector_xyz_euler(rpy)"

        This determines if the output vector will represent orientation as a quaternion or as roll-pitch-yaw Euler angles.
        If the output vector is in quaternion format, then the output will be 7-dimensional.
        If the output vector is in Euler angle format, then the output will be 6-dimensional.

        By default, this is set to "vector_xyz_quat(wxyz)".
    """
    name: str = "rigid_transform_to_vector_system"
    output_format: str = "vector_xyz_quat(wxyz)" # options: "vector_xyz_quat(wxyz)", "vector_xyz_euler(rpy)"


from .configuration import Configuration as PuppetmakerConfiguration
from .puppetmaker import Puppetmaker
from .puppet_signature import PuppetSignature

__all__ = [
    "Puppetmaker",
    "PuppetmakerConfiguration",
    "PuppetSignature",
]


from enum import IntEnum

class RollPitchYawAngle(IntEnum):
    kRoll = 0
    kPitch = 1
    kYaw = 2

    def __str__(self) -> str:
        """
        *Description*
        
        This method returns a string representation of the RollPitchYawAngle enum.
        """
        if self == RollPitchYawAngle.kRoll:
            return "Roll"
        elif self == RollPitchYawAngle.kPitch:
            return "Pitch"
        elif self == RollPitchYawAngle.kYaw:
            return "Yaw"
        else:
            raise ValueError(f"Invalid RollPitchYawAngle value: {self}")

def rpy_equivalent_body_rotation_order() -> list[RollPitchYawAngle]:
    """
    *Description*
    
    This method returns the equivalent body rotation order for Roll-Pitch-Yaw angles.
    """
    return [
        RollPitchYawAngle.kYaw,
        RollPitchYawAngle.kPitch,
        RollPitchYawAngle.kRoll,
    ]


from brom_drake.utils.leaf_systems.flexible_port_switch import FlexiblePortSwitch
from brom_drake.control.grippers import GripperController
from brom_drake.utils.leaf_systems.network_fsm import NetworkXFSM

__all__ = [
    "FlexiblePortSwitch",
    "GripperController",
    "NetworkXFSM",
]


from .config import Configuration
from .phases import AttemptGraspWithPuppeteerWristPhase
from .production import AttemptGraspWithPuppeteerWrist
from .script import Script

__all__ = [
    "AttemptGraspWithPuppeteerWristPhase",
    "Configuration",
    "Script",
    "AttemptGraspWithPuppeteerWrist",
]


from .box import BoxDefinition
from .cylinder import CylinderDefinition
from .shape_definition import ShapeDefinition, ShapeEnum
from .sphere import SphereDefinition

__all__ = [
    "BoxDefinition",
    "CylinderDefinition",
    "ShapeDefinition", "ShapeEnum",
    "SphereDefinition",
]


from brom_drake.file_manipulation.urdf.drake_ready_urdf_converter.converter import DrakeReadyURDFConverter
from brom_drake.file_manipulation.urdf.drake_ready_urdf_converter.config import (
    DrakeReadyURDFConverterConfig,
    MeshReplacementStrategy,
    MeshReplacementStrategies,
)
from .drakeify import drakeify_my_urdf
from .drake_ready_urdf_converter.util import (
    find_mesh_file_path_in,
    tree_contains_transmission_for_joint,
    create_transmission_element_for_joint,
)
from brom_drake.file_manipulation.urdf.simple_writer.urdf_definition import SimpleShapeURDFDefinition

__all__ = [
    'DrakeReadyURDFConverter',
    'DrakeReadyURDFConverterConfig',
    'MeshReplacementStrategies',
    'MeshReplacementStrategy',
    'drakeify_my_urdf',
    'find_mesh_file_path_in',
    'tree_contains_transmission_for_joint',
    'create_transmission_element_for_joint',
    'SimpleShapeURDFDefinition',
]